"""
Threading utilities.

Currently most activities are performed on the same "foreground thread".
This thread is responsible for:
(1) running the GUI and
(2) mediating access to model elements (including the underlying database).
"""

from collections import deque
from collections.abc import Callable, Generator
from crystal.util.bulkheads import (
    capture_crashes_to_stderr, ensure_is_bulkhead_call,
)
from crystal.util.profile import create_profiled_callable
from enum import Enum
from functools import partial, wraps
import os
import sys
import threading
import traceback
from typing import cast, Deque, Optional, TypeVar
from typing_extensions import ParamSpec
import wx

# Whether to explicitly manage the queue of deferred foreground tasks,
# instead of letting wxPython manage the queue internally.
# 
# Can be useful for debugging, because when True the queue of deferred foreground tasks
# is directly inspectable by examining the "_fg_calls" variable in this module.
_DEBUG_FG_CALL_QUEUE = False

# If True, then the runtime of foreground tasks is tracked to ensure
# they are short. This is necessary to keep the UI responsive.
_PROFILE_FG_TASKS = os.environ.get('CRYSTAL_NO_PROFILE_FG_TASKS', 'False') != 'True'

# Maximum reasonable time that foreground tasks should take to complete.
# If profiling is enabled, warnings will be printed for tasks whose runtime
# exceeds this threshold.
_FG_TASK_RUNTIME_THRESHOLD = 1.0 # sec

# Whether to enforce that callables scheduled with fg_call_later()
# be decorated with @capture_crashes_to*.
_DEFERRED_FG_CALLS_MUST_CAPTURE_CRASHES = True
# Whether to enforce that callables scheduled with bg_call_later()
# be decorated with @capture_crashes_to*.
_DEFERRED_BG_CALLS_MUST_CAPTURE_CRASHES = True


_P = ParamSpec('_P')
_R = TypeVar('_R')


# ------------------------------------------------------------------------------
# Access Foreground Thread

_fg_thread = None  # type: Optional[threading.Thread]


def set_foreground_thread(fg_thread: threading.Thread | None) -> None:
    global _fg_thread
    _fg_thread = fg_thread


def is_foreground_thread() -> bool:
    """
    Returns whether the current thread is the foreground thread
    (with the active wx.App object).
    """
    # NOTE: It has been observed that wx.IsMainThread() is unreliable,
    #       returning True for all threads, when there is no running main loop.
    #       Therefore Crystal keeps track of the main thread itself.
    return threading.current_thread() == _fg_thread


def has_foreground_thread() -> bool:
    """
    Returns whether any foreground thread exists.
    
    If it doesn't exist, it may not have been created yet or it may have exited.
    """
    return _fg_thread is not None


# ------------------------------------------------------------------------------
# Thread Affinity

def fg_affinity(func: Callable[_P, _R]) -> Callable[_P, _R]:
    """
    Marks the decorated function as needing to be called from the
    foreground thread only.
    
    Calling the decorated function from an inappropriate thread will immediately
    raise an AssertionError.
    
    The following kinds of manipulations need to happen on the foreground thread:
    - wxPython calls, except for wx.CallAfter
    - SQLite calls
    """
    if __debug__:  # no -O passed on command line?
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not is_foreground_thread() and has_foreground_thread():
                raise AssertionError(
                    f'fg_affinity: Expected call on foreground thread: {func}')
            return func(*args, **kwargs)  # cr-traceback: ignore
        return wrapper
    else:
        return func


def bg_affinity(func: Callable[_P, _R]) -> Callable[_P, _R]:
    """
    Marks the decorated function as needing to be called from a
    background thread only, and in particular not from the foreground thread.
    
    Calling the decorated function from an inappropriate thread will immediately
    raise an AssertionError.
    
    The following kinds of manipulations need to happen on background threads:
    - Blocking I/O (except for database I/O)
    - Long-running tasks that would block the UI if run on the foreground thread
    """
    if __debug__:  # no -O passed on command line?
        @wraps(func)
        def wrapper(*args, **kwargs):
            assert not is_foreground_thread()
            return func(*args, **kwargs)  # cr-traceback: ignore
        return wrapper
    else:
        return func


# ------------------------------------------------------------------------------
# Call on Foreground Thread

def fg_call_later(
        callable: Callable[_P, None],
        # TODO: Give `args` the type `_P` once that can be spelled in Python's type system
        *, args=(),
        profile: bool=True,
        force_later: bool=False,
        ) -> None:
    """
    Schedules the specified callable to be called on the foreground thread,
    either immediately if the caller is already running on the foreground thread,
    or later if the caller is running on a different thread.
    
    Background threads should use this method when accessing the UI or model.
    
    Arguments:
    * callable -- the callable to run.
    * args -- the arguments to provide to the callable.
    * profile -- 
        whether to profile the callable's runtime.
        True by default so that warnings are printed for slow foreground tasks.
    * force_later --
        whether to force scheduling the callable later, even if the caller is
        already running on the foreground thread.
    
    Raises:
    * NoForegroundThreadError
    """
    if not has_foreground_thread():
        raise NoForegroundThreadError()
    
    is_fg_thread = is_foreground_thread()  # cache
    
    if _DEFERRED_FG_CALLS_MUST_CAPTURE_CRASHES:
        ensure_is_bulkhead_call(callable)
    
    if _PROFILE_FG_TASKS and profile and not is_fg_thread:
        callable = create_profiled_callable(
            'Slow foreground task',
            _FG_TASK_RUNTIME_THRESHOLD,
            callable, *args
        )
        args=()
    
    if is_fg_thread and not force_later:
        callable(*args)
    else:
        try:
            if _DEBUG_FG_CALL_QUEUE:
                if len(args) == 0:
                    _fg_call_later_in_order(callable)
                else:
                    _fg_call_later_in_order(partial(callable, *args))
            else:
                wx.CallAfter(callable, *args)
        except Exception as e:
            if not has_foreground_thread():
                raise NoForegroundThreadError()
            
            # ex: RuntimeError: wrapped C/C++ object of type PyApp has been deleted
            if str(e) == 'wrapped C/C++ object of type PyApp has been deleted':
                raise NoForegroundThreadError()
            # ex: AssertionError: No wx.App created yet
            elif str(e) == 'No wx.App created yet':
                raise NoForegroundThreadError()
            else:
                raise


_fg_calls = deque()  # type: Deque[Callable[[], None]]

def _fg_call_later_in_order(callable: Callable[[], None]) -> None:
    _fg_calls.append(callable)
    wx.CallAfter(_do_fg_calls)

def _do_fg_calls() -> None:
    while True:
        try:
            cur_fg_call = _fg_calls.popleft()  # type: Callable[[], None]
        except IndexError:
            break
        try:
            cur_fg_call()
        except Exception:
            # Print traceback of any unhandled exception
            traceback.print_exc()


def fg_call_and_wait(
        callable: Callable[_P, _R],
        # TODO: Give `args` the type `_P` once that can be spelled in Python's type system
        *, args=(),
        profile: bool=True
        ) -> _R:
    """
    Calls the specified callable on the foreground thread and waits for it to complete,
    returning the result of the callable, including any raised exception.

    Background threads should use this method when accessing the UI or model.
    
    Arguments:
    * callable -- the callable to run.
    * args -- the arguments to provide to the callable.
    * profile -- 
        whether to profile the callable's runtime.
        True by default so that warnings are printed for slow foreground tasks.
    
    Raises:
    * NoForegroundThreadError
    """
    if not has_foreground_thread():
        raise NoForegroundThreadError()
    
    if is_foreground_thread():
        return callable(*args)  # cr-traceback: ignore
    else:
        event = threading.Event()
        callable_started = False
        callable_result = None
        callable_exc_info = None
        
        waiting_calling_thread = threading.current_thread()  # capture
        
        @capture_crashes_to_stderr
        def fg_task() -> None:
            nonlocal callable_started, callable_result, callable_exc_info
            
            callable_started = True
            
            # Run task
            fg_thread = threading.current_thread()
            setattr(fg_thread, '_cr_waiting_calling_thread', waiting_calling_thread)
            try:
                callable_result = callable(*args)  # cr-traceback: ignore
            except BaseException as e:
                callable_exc_info = sys.exc_info()
            finally:
                setattr(fg_thread, '_cr_waiting_calling_thread', None)
            
            # Send signal
            event.set()
        fg_task.callable = callable  # type: ignore[attr-defined]
        fg_call_later(fg_task, profile=profile)
        
        # Wait for signal
        while True:
            if not callable_started:
                if event.wait(timeout=1.0):
                    break
                if not has_foreground_thread():
                    # Presumably the foreground thread did shutdown
                    # while fg_call_later was scheduling a callable
                    # that now will never actually run
                    raise NoForegroundThreadError()
            else:
                if event.wait():
                    break
        
        # Reraise callable's exception, if applicable
        if callable_exc_info is not None:
            exc_info = callable_exc_info
            assert exc_info[1] is not None
            raise exc_info[1].with_traceback(exc_info[2])  # cr-traceback: ignore
        
        return cast(_R, callable_result)


class NoForegroundThreadError(ValueError):
    pass


@fg_affinity
def fg_waiting_calling_thread() -> threading.Thread | None:
    """
    If the current task running on the foreground thread was scheduled
    by a call to fg_call_and_wait(), returns the thread that made that call.
    Otherwise returns None.
    """
    fg_thread = threading.current_thread()
    return getattr(fg_thread, '_cr_waiting_calling_thread', None)


# ------------------------------------------------------------------------------
# Call on Background Thread

def bg_call_later(
        callable: Callable[_P, None],
        # TODO: Give `args` the type `_P` once that can be spelled in Python's type system
        *, args=(),
        daemon: bool=False,
        ) -> threading.Thread:
    """
    Calls the specified callable on a new background thread.
    
    Arguments:
    * callable -- the callable to run.
    * args -- the arguments to provide to the callable.
    * daemon -- 
        if True, forces the background thread to be a daemon,
        and not prevent program termination while it is running.
    """
    if _DEFERRED_BG_CALLS_MUST_CAPTURE_CRASHES:
        ensure_is_bulkhead_call(callable)
    thread = threading.Thread(target=callable, args=args, daemon=daemon)
    thread.start()
    return thread


# ------------------------------------------------------------------------------
# Thread Switching Coroutines

class SwitchToThread(Enum):
    BACKGROUND = 1
    FOREGROUND = 2


def start_thread_switching_coroutine(
        first_command: SwitchToThread,
        coro: Generator[SwitchToThread, None, _R],
        capture_crashes_to_deco: Callable[[Callable[[], None]], Callable[[], None]],
        /,
        ) -> None:
    """
    Starts the specified thread-switching-coroutine on a new background thread.
    
    A thread-switching-coroutine starts executing on the thread given by `first_command`.
    Whenever the coroutine wants to switch to a different thread it yields
    a new `SwitchToThread` command. A coroutine may ask to switch to the
    same thread it was running on before without any effect.
    
    If the caller is running on the foreground thread and the prefix of the
    thread-switching-coroutine is configured to execute on the foreground thread
    (when first_command == SwitchToThread.FOREGROUND and subsequent yields
    also say SwitchToThread.FOREGROUND), then that prefix is run to completion
    synchronously before this method returns.
    """
    # If is foreground thread, immediately run any prefix of the coroutine
    # that wants to be on the foreground thread
    if is_foreground_thread():
        @capture_crashes_to_deco
        def fg_task() -> None:
            nonlocal first_command
            
            command = first_command
            while command == SwitchToThread.FOREGROUND:
                try:
                    command = next(coro)
                except StopIteration as e:
                    return
            assert command == SwitchToThread.BACKGROUND
            
            first_command = command  # reinterpret
        fg_task()
    
    # Run remainder of coroutine on a new background thread
    @capture_crashes_to_deco
    def bg_task() -> None:
        run_thread_switching_coroutine(first_command, coro)
    bg_call_later(bg_task)


@bg_affinity
def run_thread_switching_coroutine(
        first_command: SwitchToThread,
        coro: Generator[SwitchToThread, None, _R],
        /,
        ) -> _R:
    """
    Runs the specified thread-switching-coroutine on the caller's
    background thread until it completes.
    
    A thread-switching-coroutine starts executing on the thread given by `first_command`.
    Whenever the coroutine wants to switch to a different thread it yields
    a new `SwitchToThread` command. A coroutine may ask to switch to the
    same thread it was running on before without any effect.
    
    Raises if the caller is not running on a background thread.
    """
    command = first_command
    while True:
        if command == SwitchToThread.BACKGROUND:
            run_next = _CALL_NOW
        elif command == SwitchToThread.FOREGROUND:
            run_next = fg_call_and_wait
        else:
            raise AssertionError()
        try:
            command = run_next(lambda: next(coro))  # cr-traceback: ignore
        except StopIteration as e:
            return e.value


_CALL_NOW = lambda f: f()


# ------------------------------------------------------------------------------
# Quitting

_is_quitting = False


def is_quitting() -> bool:
    return _is_quitting


def set_is_quitting() -> None:
    global _is_quitting
    _is_quitting = True


# ------------------------------------------------------------------------------



from collections.abc import Iterator
from contextlib import contextmanager
from crystal.model import Project
from crystal.tests.util.skip import skipTest
from crystal.tests.util.subtests import awith_subtests, SubtestsContext
from crystal.tests.util.windows import OpenOrCreateDialog
from crystal.util.xos import is_mac_os, is_windows
import os
import stat
import subprocess
import tempfile
from unittest import skip

# ------------------------------------------------------------------------------
# Tests

async def test_project_opens_as_readonly_when_user_requests_it_in_ui() -> None:
    with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
        # Create empty project
        async with (await OpenOrCreateDialog.wait_for()).create(project_dirpath) as (mw, _):
            pass
        
        # Ensure project opens as writable by default
        async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
            assert False == mw.readonly
        
        # Ensure project opens as readonly when requested
        async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath, readonly=True) as (mw, project):
            assert True == mw.readonly


@skip('not yet automated')
async def test_project_opens_as_readonly_when_user_requests_it_in_cli() -> None:
    pass


async def test_project_opens_as_readonly_when_project_is_on_readonly_filesystem() -> None:
    if not is_mac_os():
        skipTest('only supported on macOS')
    
    with tempfile.TemporaryDirectory() as working_dirpath:
        volume_src_dirpath = os.path.join(working_dirpath, 'Project')
        os.mkdir(volume_src_dirpath)
        
        # Create empty project
        project_dirpath = os.path.join(volume_src_dirpath, 'Project.crystalproj')
        async with (await OpenOrCreateDialog.wait_for()).create(project_dirpath) as (mw, _):
            pass
        
        # Create disk image with project on it
        dmg_filepath = os.path.join(working_dirpath, 'Project.dmg')
        subprocess.run([
            'hdiutil',
            'create',
            '-srcfolder', volume_src_dirpath,
            dmg_filepath,
        ], check=True, stdout=subprocess.DEVNULL)
        
        # Mount disk image as readonly
        subprocess.run([
            'hdiutil',
            'attach',
            dmg_filepath,
            '-readonly',
        ], check=True, stdout=subprocess.DEVNULL)
        volume_dirpath = '/Volumes/Project'
        assert os.path.exists(volume_dirpath)
        try:
            mounted_project_dirpath = os.path.join(volume_dirpath, 'Project.crystalproj')
            async with (await OpenOrCreateDialog.wait_for()).open(mounted_project_dirpath) as (mw, project):
                assert True == mw.readonly
        finally:
            subprocess.run([
                'umount',
                volume_dirpath,
            ], check=True)


@awith_subtests
async def test_project_opens_as_readonly_when_project_directory_or_database_is_locked(subtests: SubtestsContext) -> None:
    with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
        # Create empty project
        async with (await OpenOrCreateDialog.wait_for()).create(project_dirpath) as (mw, _):
            pass
        
        # Ensure project opens as writable by default, when no files are locked
        with subtests.test(locked='nothing'):
            async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
                assert False == mw.readonly
        
        # Ensure project opens as readonly when project directory is locked
        if is_windows():
            # Can't lock directory on Windows
            pass
        else:
            with subtests.test(locked='project_directory'):
                with _file_set_to_readonly(project_dirpath):
                    async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
                        assert True == mw.readonly
        
        # Ensure project opens as readonly when project database is locked
        with subtests.test(locked='project_database'):
            with _file_set_to_readonly(os.path.join(project_dirpath, Project._DB_FILENAME)):
                async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
                    assert True == mw.readonly


# ------------------------------------------------------------------------------
# Utility

@contextmanager
def _file_set_to_readonly(filepath: str) -> Iterator[None]:
    _set_file_readonly(filepath, True)
    try:
        yield
    finally:
        _set_file_readonly(filepath, False)


def _set_file_readonly(filepath: str, readonly: bool) -> None:
    if is_mac_os():
        # Set the "Locked" attribute
        flags = os.stat(filepath).st_flags  # type: ignore[attr-defined]  # available on macOS
        if readonly:
            os.chflags(filepath, flags | stat.UF_IMMUTABLE)  # type: ignore[attr-defined]  # available on macOS
        else:
            os.chflags(filepath, flags & ~stat.UF_IMMUTABLE)  # type: ignore[attr-defined]  # available on macOS
    else:
        # Set the "Read Only" attribute on Windows,
        # or set the presence of "owner writer" permissions on Linux
        mode = os.stat(filepath).st_mode
        if readonly:
            os.chmod(filepath, mode & ~stat.S_IWRITE)
        else:
            os.chmod(filepath, mode | stat.S_IWRITE)


# ------------------------------------------------------------------------------



"""
HTML parser implementation that uses regular expressions.
"""

from __future__ import annotations

from crystal.doc.generic import Document, Link
import re


def parse_html_and_links(
        html_bytes: bytes, 
        declared_charset: str | None=None
        ) -> tuple[Document, list[Link]]:
    if not isinstance(html_bytes, bytes):
        raise ValueError('This parser implementation only accepts bytestrings.')
    # TODO: Attempt to honor the declared_charset rather than always assuming UTF-8.
    #       Recommend using the webencodings package (already a dependency) to map
    #       internet charset names to Python encoding names.
    html = html_bytes.decode('utf-8', errors='replace')
    
    dividers_and_urls = re.split(r'(?i)([\'"][^\'"]+\.s?html?[\'"])', html)
    
    dividers_and_links = []
    links = []  # type: list[Link]
    for (i, old_item) in enumerate(dividers_and_urls):
        if i & 1 == 0:
            new_item = old_item
        else:
            new_item = BasicLink(old_item)
            links.append(new_item)
        
        dividers_and_links.append(new_item)
    
    return (BasicDocument(dividers_and_links), links)


class BasicDocument(Document):
    def __init__(self, dividers_and_links: list[str | BasicLink]) -> None:
        self._dividers_and_links = dividers_and_links
    
    def __str__(self) -> str:
        return ''.join([str(item) for item in self._dividers_and_links])


class BasicLink(Link):
    def __init__(self, quoted_href: str) -> None:
        if not (len(quoted_href) >= 2 and 
                quoted_href[0] in ('\'', '\"') and 
                quoted_href[-1] == quoted_href[0]):
            raise ValueError()
        self._quoted_href = quoted_href
        
        self.title = None
        self.type_title = 'Unknown'
        self.embedded = False
    
    def _get_relative_url(self) -> str:
        return self._quoted_href[1:-1]
    def _set_relative_url(self, value: str) -> None:
        self._quoted_href = '"%s"' % value
    relative_url = property(_get_relative_url, _set_relative_url)
    
    # Simplifies BasicDocument's __str__ method implementation
    def __str__(self) -> str:
        return self._quoted_href



"""Tests for DownloadResourceBodyTask"""

from collections.abc import Callable, Iterable, Iterator
from contextlib import contextmanager
from crystal import resources
from crystal.model import (
    Project, ProjectHasTooManyRevisionsError, Resource, ResourceRevision,
)
from crystal.tests.util.runner import bg_sleep
from crystal.tests.util.server import served_project
from crystal.tests.util.wait import DEFAULT_WAIT_PERIOD
from crystal.tests.util.windows import OpenOrCreateDialog
from crystal.util.db import DatabaseCursor
import errno
from http.client import HTTPConnection, HTTPResponse
from http.server import BaseHTTPRequestHandler, HTTPServer
import os
import sqlite3
import tempfile
import threading
from typing import BinaryIO, cast, NoReturn, Self
from unittest import skip
from unittest.mock import ANY, Mock, patch

# ------------------------------------------------------------------------------
# Tests: Success Cases

async def test_download_does_save_resource_metadata_and_content_accurately() -> None:
    HEADERS = [
        ['Content-Type', 'application/zip'],
        ['Server', 'TestServer/1.0'],
        ['Date', 'Wed, 16 Aug 2023 00:00:00 GMT'],  # arbitrary
    ]
    
    with resources.open_binary('testdata_xkcd.crystalproj.zip') as content_file:
        content_bytes = content_file.read()
    assert len(content_bytes) > 1000  # ensure is a reasonably sized file
    
    with _file_served(HEADERS, content_bytes) as server_port:
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            r = Resource(project, f'http://localhost:{server_port}/')
            revision_future = r.download_body()
            while not revision_future.done():
                await bg_sleep(DEFAULT_WAIT_PERIOD)
            
            downloaded_revision = revision_future.result()  # type: ResourceRevision
            loaded_revision = r.default_revision()
            assert loaded_revision is not None
            
            for revision in [downloaded_revision, loaded_revision]:
                # Ensure no error is reported
                assert None == revision.error
                
                # Ensure metadata is correct
                EXPECTED_METADATA = {
                    'http_version': 10,  # HTTP/1.0
                    'status_code': 200,
                    'reason_phrase': 'OK',
                    'headers': HEADERS,
                }
                assert EXPECTED_METADATA == revision.metadata
                
                # Ensure content is correct
                with revision.open() as saved_content:
                    saved_content_bytes = saved_content.read()
                assert content_bytes == saved_content_bytes


async def test_download_does_autopopulate_date_header_if_not_received_from_origin() -> None:
    HEADERS = [
        ['Content-Type', 'application/zip'],
        ['Server', 'TestServer/1.0'],
        # (No 'Date' header)
    ]
    EXPECTED_HEADERS_SAVED = [
        ['Content-Type', 'application/zip'],
        ['Server', 'TestServer/1.0'],
        ['Date', ANY],
    ]
    
    with resources.open_binary('testdata_xkcd.crystalproj.zip') as content_file:
        content_bytes = content_file.read()
    
    with _file_served(HEADERS, content_bytes) as server_port:
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            r = Resource(project, f'http://localhost:{server_port}/')
            revision_future = r.download_body()
            while not revision_future.done():
                await bg_sleep(DEFAULT_WAIT_PERIOD)
            
            downloaded_revision = revision_future.result()  # type: ResourceRevision
            loaded_revision = r.default_revision()
            assert loaded_revision is not None
            
            for revision in [downloaded_revision, loaded_revision]:
                EXPECTED_METADATA = {
                    'http_version': 10,  # HTTP/1.0
                    'status_code': 200,
                    'reason_phrase': 'OK',
                    'headers': EXPECTED_HEADERS_SAVED,
                }
                assert EXPECTED_METADATA == revision.metadata
                
                assert None != revision.date, \
                    'Date header has invalid value'


@contextmanager
def _file_served(headers: list[list[str]], content_bytes: bytes) -> Iterator[int]:
    class RequestHandler(BaseHTTPRequestHandler):
        def do_GET(self) -> None:  # override
            self.send_response_only(200)
            for (k, v) in headers:
                self.send_header(k, v)
            self.end_headers()
            self.wfile.write(content_bytes)
    
    # Start an HTTP server that serves a test file for any GET request
    with HTTPServer(('', 0), RequestHandler) as server:
        (_, server_port) = server.server_address
        
        def do_serve_forever() -> None:
            try:
                server.serve_forever()
            except OSError as e:
                # [WinError 10038] An operation was attempted on something that is not a socket
                if getattr(e, 'winerror', None) == 10038:
                    # Already called server_close(). Ignore error.
                    pass
                else:
                    raise
        server_thread = threading.Thread(target=do_serve_forever, daemon=True)
        server_thread.start()
        try:
            yield server_port
        finally:
            server.server_close()


# ------------------------------------------------------------------------------
# Tests: Error Cases

async def test_when_no_errors_then_database_row_and_body_file_is_created_and_returns_normal_revision() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            r = Resource(project, home_url)
            revision_future = r.download_body()
            while not revision_future.done():
                await bg_sleep(DEFAULT_WAIT_PERIOD)
            
            # Ensure no error is reported
            revision = revision_future.result()  # type: ResourceRevision
            assert None == revision.error
            
            # Ensure database and filesystem is in expected state
            assert 1 == project._revision_count()
            assert True == os.path.exists(os.path.join(
                project.path, Project._REVISIONS_DIRNAME,
                '000', '000', '000', '000', f'{revision._id:03x}'))
            assert [] == os.listdir(os.path.join(
                project.path, Project._TEMPORARY_DIRNAME))


async def test_when_network_io_error_then_tries_to_delete_partial_body_file_but_leave_database_row_and_returns_error_revision() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with _downloads_mocked_to_raise_network_io_error() as is_connection_reset:
                r = Resource(project, home_url)
                revision_future = r.download_body()
                while not revision_future.done():
                    await bg_sleep(DEFAULT_WAIT_PERIOD)
            
            # Ensure an I/O error is reported as an error revision
            revision = revision_future.result()  # type: ResourceRevision
            assert None != revision.error
            assert is_connection_reset(revision.error)
            
            # Ensure database and filesystem is in expected state
            assert 1 == project._revision_count()
            assert False == os.path.exists(revision._body_filepath)
            assert [] == os.listdir(os.path.join(
                project.path, Project._TEMPORARY_DIRNAME))


async def test_when_database_error_then_tries_to_delete_partial_body_file_and_raises_database_error() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with _database_cursor_mocked_to_raise_database_io_error_on_write(project) as is_database_error:
                r = Resource(project, home_url)
                revision_future = r.download_body()
                while not revision_future.done():
                    await bg_sleep(DEFAULT_WAIT_PERIOD)
            
            # Ensure a database error is reported as a raised exception
            try:
                revision = revision_future.result()  # type: ResourceRevision
            except Exception as e:
                assert is_database_error(e)
            else:
                assert False, 'Expected database error'
            
            # Ensure database and filesystem is in expected state
            assert 0 == project._revision_count()
            assert [] == os.listdir(os.path.join(
                project.path, Project._REVISIONS_DIRNAME))
            assert [] == os.listdir(os.path.join(
                project.path, Project._TEMPORARY_DIRNAME))


async def test_when_network_io_error_and_database_error_then_tries_to_delete_partial_body_file_and_raises_database_error() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with _downloads_mocked_to_raise_network_io_error() as is_connection_reset:
                with _database_cursor_mocked_to_raise_database_io_error_on_write(project) as is_database_error:
                    r = Resource(project, home_url)
                    revision_future = r.download_body()
                    while not revision_future.done():
                        await bg_sleep(DEFAULT_WAIT_PERIOD)
                
            # Ensure a database error is reported as a raised exception
            # (rather than reporting the I/O error as an error revision)
            try:
                revision = revision_future.result()  # type: ResourceRevision
            except Exception as e:
                assert is_database_error(e)
            else:
                assert False, 'Expected database error'
            
            # Ensure database and filesystem is in expected state
            assert 0 == project._revision_count()
            assert [] == os.listdir(os.path.join(
                project.path, Project._REVISIONS_DIRNAME))
            assert [] == os.listdir(os.path.join(
                project.path, Project._TEMPORARY_DIRNAME))


async def test_when_open_project_given_partial_body_files_exist_then_deletes_all_partial_body_files() -> None:
    with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
        # Create empty project
        async with (await OpenOrCreateDialog.wait_for()).create(project_dirpath) as (mw, _):
            pass
        assert [] == os.listdir(os.path.join(
            project_dirpath, Project._TEMPORARY_DIRNAME))
        
        # Insert a partial body file manually
        with open(os.path.join(project_dirpath, Project._TEMPORARY_DIRNAME, '1.body'), 'wb') as f:
            pass
        assert [] != os.listdir(os.path.join(
            project_dirpath, Project._TEMPORARY_DIRNAME))
        
        # Reopen project
        async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
            # Ensure partial body file is deleted
            assert [] == os.listdir(os.path.join(
                project_dirpath, Project._TEMPORARY_DIRNAME))


# ------------------------------------------------------------------------------
# Tests: Specific Error Cases

@skip('covered by: test_when_network_io_error_then_tries_to_delete_partial_body_file_but_leave_database_row_and_returns_error_revision')
async def test_given_downloading_revision_when_reading_from_network_raises_io_error_then_returns_error_revision() -> None:
    pass


async def test_given_downloading_revision_when_writing_to_disk_raises_io_error_then_returns_error_revision() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with _downloads_mocked_to_raise_disk_io_error() as is_io_error:
                r = Resource(project, home_url)
                revision_future = r.download_body()
                while not revision_future.done():
                    await bg_sleep(DEFAULT_WAIT_PERIOD)
            
            # Ensure an I/O error is reported as an error revision
            revision = revision_future.result()  # type: ResourceRevision
            assert None != revision.error
            assert is_io_error(revision.error)
            
            # Ensure database and filesystem is in expected state
            assert 1 == project._revision_count()
            assert False == os.path.exists(revision._body_filepath)
            assert [] == os.listdir(os.path.join(
                project.path, Project._TEMPORARY_DIRNAME))


async def test_given_downloading_revision_when_writing_to_disk_raises_io_error_and_writing_to_database_raises_io_error_then_raises_database_error() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with _downloads_mocked_to_raise_disk_io_error() as is_io_error:
                with _database_cursor_mocked_to_raise_database_io_error_on_write(project) as is_database_error:
                    r = Resource(project, home_url)
                    revision_future = r.download_body()
                    while not revision_future.done():
                        await bg_sleep(DEFAULT_WAIT_PERIOD)
            
            # Ensure a database error is reported as a raised exception
            try:
                revision = revision_future.result()  # type: ResourceRevision
            except Exception as e:
                assert is_database_error(e)
            else:
                assert False, 'Expected database error'
            
            # Ensure database and filesystem is in expected state
            assert 0 == project._revision_count()
            assert [] == os.listdir(os.path.join(
                project.path, Project._REVISIONS_DIRNAME))
            assert [] == os.listdir(os.path.join(
                project.path, Project._TEMPORARY_DIRNAME))


# NOTE: This error scenario is expected to never happen in practice,
#       so there's limited value in optimizing handling for it.
# TODO: Consider alter behavior to still return an error revision,
#       but NOT persist the error revision to disk.
# TODO: Also report the I/O error to the UI in some fashion rather
#       than silently dropping it.
async def test_given_project_has_revision_with_maximum_id_when_download_revision_then_raises_error_and_error_revision_not_created() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        home_url = sp.get_request_url('https://xkcd.com/')
    
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            old_revision_count = project._revision_count()  # capture
            
            resource = Resource(project, home_url)
            
            with _database_cursor_mocked_to_create_revision_with_id(Project._MAX_REVISION_ID + 1, project):
                download_result = resource.download_body()
                while not download_result.done():
                    await bg_sleep(DEFAULT_WAIT_PERIOD)
                try:
                    download_result.result()
                except ProjectHasTooManyRevisionsError:
                    pass
                else:
                    raise AssertionError(
                        'Expected ProjectHasTooManyRevisionsError to be raised')
            
            new_revision_count = project._revision_count()
            assert old_revision_count == new_revision_count


# ------------------------------------------------------------------------------
# Utility

@contextmanager
def _downloads_mocked_to_raise_network_io_error() -> Iterator:
    def raise_connection_reset() -> NoReturn:
        raise OSError('Connection reset')
    def is_connection_reset(e: object) -> bool:
        return (
            isinstance(e, IOError) and
            str(e) == 'Connection reset'
        )
    
    DATA_BEFORE_EOF = b'<html>'
    
    def mock_read() -> Iterable[bytes]:
        yield DATA_BEFORE_EOF
        raise_connection_reset()
    def mock_readinto() -> Iterable[int]:
        yield len(DATA_BEFORE_EOF)
        raise_connection_reset()
    
    mock_response = Mock(spec=HTTPResponse)
    mock_response.version = 10  # HTTP/1.0
    mock_response.status = 200
    mock_response.reason = 'OK'
    mock_response.getheaders = Mock(return_value=[
        ('Content-Type', 'text/html; charset=utf-8')
    ])
    mock_response.read = Mock(side_effect=mock_read())
    mock_response.readinto = Mock(side_effect=mock_readinto())
    mock_response.fileno = Mock(side_effect=OSError)  # no file descriptor
    
    mock_connection = Mock(spec_set=HTTPConnection)
    mock_connection.getresponse = Mock(return_value=mock_response)
    mock_connection.close = Mock(return_value=None)
    with patch(
            # HTTPConnection used by HttpResourceRequest.__call__()
            'crystal.download.HTTPConnection',
            return_value=mock_connection):
        
        # Shrink the copy buffer size in the "length" parameter,
        # because Python 3.12+ assumes that a read that doesn't fill the
        # copy buffer is an EOF, without checking. We want to ensure that
        # at least 2 calls to read() or readinto() are made.
        from crystal.util import xshutil
        super_copyfileobj_readinto = xshutil.copyfileobj_readinto
        def copyfileobj_readinto(src, dst, length=len(DATA_BEFORE_EOF)):
            return super_copyfileobj_readinto(src, dst, length=length)
        with patch(
                # xshutil.copyfileobj_readinto used by ResourceRevision._create_from_stream()
                'crystal.util.xshutil.copyfileobj_readinto',
                copyfileobj_readinto):
            yield is_connection_reset


@contextmanager
def _downloads_mocked_to_raise_disk_io_error() -> Iterator[Callable[[object], bool]]:
    def fake_write(b: bytes) -> None:
        # Simulate faulty disk
        raise OSError(errno.EIO, 'Input/output error')
    def is_io_error(e: object) -> bool:
        return (
            isinstance(e, OSError) and
            e.errno == errno.EIO
        )
    
    def FakeNamedTemporaryFile(*args, **kwargs) -> BinaryIO:
        f = tempfile.NamedTemporaryFile(*args, **kwargs)
        if kwargs.get('suffix', '').endswith('.body'):
            f.write = fake_write  # type: ignore[assignment]
        return cast(BinaryIO, f)
    
    with patch('crystal.model.NamedTemporaryFile', FakeNamedTemporaryFile):
        yield is_io_error


@contextmanager
def _database_cursor_mocked_to_raise_database_io_error_on_write(
        project: Project
        ) -> Iterator[Callable[[object], bool]]:
    def raise_database_error() -> NoReturn:
        raise sqlite3.OperationalError('database is locked')
    def is_database_error(e: object) -> bool:
        return (
            isinstance(e, sqlite3.OperationalError) and
            str(e) == 'database is locked'
        )
    
    real_cursor_func = project._db.cursor  # capture
    def mock_cursor():
        real_cursor = real_cursor_func()
        
        def mock_execute(command: str, *args, **kwargs):
            if command.startswith('insert into resource_revision '):
                raise_database_error()
            else:
                return real_cursor.execute(command, *args, **kwargs)
        
        mock_cursor = Mock(wraps=real_cursor)
        mock_cursor.execute = mock_execute
        type(mock_cursor).lastrowid = property(lambda self: real_cursor.lastrowid)
        return mock_cursor
    
    with patch.object(project._db, 'cursor', mock_cursor):
        yield is_database_error


@contextmanager
def _database_cursor_mocked_to_create_revision_with_id(new_row_id: int, project: Project) -> Iterator[None]:
    """
    Forces any created ResourceRevision row to use the specified ID.
    """
    # Where c = project._db.cursor(),
    # mock c.execute() so that if query starts with
    # 'insert into resource_revision ' then
    # 1. followup query is made to alter the created row's
    #    ID to match a desired value
    # 2. the next call to c.lastrowid returns that new ID
    
    class FakeDatabaseCursor:  # logically extends DatabaseCursor
        def __init__(self, base) -> None:
            self._base = base
        
        def execute(self, command: str, *args, **kwargs) -> Self:
            self._base.execute(command, *args, **kwargs)
            
            did_create_revision = command.startswith(
                'insert into resource_revision ')
            if did_create_revision:
                old_row_id = self._base.lastrowid
                
                self._base.execute(
                    'update resource_revision set id = ? where id = ?',
                    (new_row_id, old_row_id))
                self._base.lastrowid = new_row_id
            
            return self
        
        def __getattr__(self, attr_name: str) -> object:
            return getattr(self._base, attr_name)
    
    real_cursor = project._db.cursor  # capture
    def fake_cursor(*args, **kwargs) -> DatabaseCursor:
        return cast(DatabaseCursor, FakeDatabaseCursor(
            real_cursor(*args, **kwargs)))
    
    with patch.object(project._db, 'cursor', fake_cursor):
        yield


# ------------------------------------------------------------------------------


"""
Tests that .crystalproj and .crystalopen icons show up correctly
in the desktop environment.
"""
from unittest import skip

# === Icon Tests for .crystalproj ===

# --- macOS ---

@skip('not yet automated')
def test_given_macos_and_did_build_app_then_crystalproj_has_custom_icon_in_finder_and_in_open_and_save_dialogs(self) -> None:
    pass


# --- Windows ---

@skip('not yet automated')
def test_given_windows_and_did_install_with_exe_then_crystalproj_has_custom_icon_in_explorer_and_in_open_and_save_dialogs(self) -> None:
    pass


# --- Linux + GNOME ---

@skip('not yet automated')
def test_given_linux_and_gnome_desktop_environment_and_did_install_to_desktop_and_crystalproj_on_desktop_then_crystalproj_has_custom_icon_on_desktop(self) -> None:
    pass


@skip('not yet automated')
def test_given_linux_and_gnome_desktop_environment_and_did_install_to_desktop_and_crystalproj_not_on_desktop_then_crystalproj_has_custom_icon_in_file_manager(self) -> None:
    pass


@skip('not yet automated')
def test_given_linux_and_gnome_desktop_environment_and_did_install_to_desktop_then_crystalproj_has_custom_icon_in_open_and_save_dialogs(self) -> None:
    pass


# --- Linux + KDE ---

@skip('not yet automated')
def test_given_linux_and_kde_desktop_environment_and_did_install_to_desktop_and_crystalproj_on_desktop_then_crystalproj_has_custom_icon_on_desktop(self) -> None:
    pass


@skip('not yet automated')
def test_given_linux_and_kde_desktop_environment_and_did_install_to_desktop_and_crystalproj_not_on_desktop_then_crystalproj_has_custom_icon_in_file_manager(self) -> None:
    pass


@skip('not yet automated')
def test_given_linux_and_kde_desktop_environment_and_did_install_to_desktop_then_crystalproj_has_custom_icon_in_open_and_save_dialogs(self) -> None:
    pass


# === Icon Tests for .crystalopen ===

# --- macOS ---

@skip('not yet automated')
def test_given_macos_and_did_build_app_then_crystalopen_has_custom_icon_in_finder_and_in_open_and_save_dialogs(self) -> None:
    pass


# --- Windows ---

@skip('not yet automated')
def test_given_windows_and_did_install_with_exe_then_crystalopen_has_custom_icon_in_explorer_and_in_open_and_save_dialogs(self) -> None:
    pass


# --- Linux + GNOME ---

@skip('not yet automated')
def test_given_linux_and_gnome_desktop_environment_and_did_install_to_desktop_and_crystalopen_not_on_desktop_then_crystalopen_has_custom_icon_in_file_manager(self) -> None:
    pass


@skip('not yet automated')
def test_given_linux_and_gnome_desktop_environment_and_did_install_to_desktop_then_crystalopen_has_custom_icon_in_open_and_save_dialogs(self) -> None:
    pass


# --- Linux + KDE ---

@skip('not yet automated')
def test_given_linux_and_kde_desktop_environment_and_did_install_to_desktop_and_crystalopen_not_on_desktop_then_crystalopen_has_custom_icon_in_file_manager(self) -> None:
    pass


@skip('not yet automated')
def test_given_linux_and_kde_desktop_environment_and_did_install_to_desktop_then_crystalopen_has_custom_icon_in_open_and_save_dialogs(self) -> None:
    pass





# NOTE: py2app doesn't appear to bundle the email.message module correctly,
#       which is used by the Python 3.11+ version of parse_header().
#       
#       Currently this doesn't matter because release builds of Crystal
#       are compiled with Python 3.8, but may matter in the future.
def parse_header(mime_header_value: str) -> tuple[str, dict]:
    """
        Parse a MIME header (such as Content-Type) into a main value and
        a dictionary of parameters.
        """
    from email.message import EmailMessage
    msg = EmailMessage()
    msg['content-type'] = mime_header_value
    return (msg.get_content_type(), msg['content-type'].params)



from contextlib import redirect_stderr
from crystal import task
from crystal.browser import MainWindow as RealMainWindow
from crystal.browser.entitytree import _ErrorNode, ResourceGroupNode
from crystal.browser.tasktree import TaskTreeNode
from crystal.doc.generic import create_external_link, Link
from crystal.model import Project, Resource, ResourceGroup, RootResource
import crystal.task
from crystal.task import (
    _DownloadResourcesPlaceholderTask, CrashedTask,
    DownloadResourceGroupMembersTask, DownloadResourceGroupTask,
    DownloadResourceTask, ParseResourceRevisionLinks, Task,
    UpdateResourceGroupMembersTask,
)
from crystal.tests.util.controls import (
    click_button, select_menuitem_now, TreeItem,
)
from crystal.tests.util.runner import pump_wx_events
from crystal.tests.util.server import served_project
from crystal.tests.util.skip import skipTest
from crystal.tests.util.tasks import (
    append_deferred_top_level_tasks, clear_top_level_tasks_on_exit,
    first_task_title_progression, MAX_DOWNLOAD_DURATION_PER_ITEM,
    scheduler_disabled, scheduler_thread_context, step_scheduler,
    step_scheduler_now, ttn_for_task,
)
from crystal.tests.util.wait import (
    first_child_of_tree_item_is_not_loading_condition, wait_for, wait_while,
)
from crystal.tests.util.windows import OpenOrCreateDialog
from crystal.tests.util.xthreading import bg_call_and_wait as _bg_call_and_wait
from crystal.ui.tree import NodeView
from crystal.util import cli
from crystal.util.bulkheads import (
    capture_crashes_to_bulkhead_arg as capture_crashes_to_task_arg,
)
from crystal.util.bulkheads import (
    capture_crashes_to_self, capture_crashes_to_stderr, crashes_captured_to,
    run_bulkhead_call,
)
from crystal.util.bulkheads import BulkheadCell, capture_crashes_to
from crystal.util.ellipsis import Ellipsis, EllipsisType
from crystal.util.wx_bind import bind
from crystal.util.xos import is_asan, is_ci, is_mac_os
from crystal.util.xthreading import (
    bg_call_later, fg_call_and_wait, is_foreground_thread,
)
from io import StringIO
import sys
import threading
from typing import List, Tuple
from unittest import skip
from unittest.mock import patch
import wx

_CRASH = ValueError('Simulated crash')


# Marks tests that frequently trigger use-after-free errors by closing
# a project that is still in the middle of running tasks.
# 
# When CRYSTAL_IGNORE_USE_AFTER_FREE=False has been made the default
# and marked tests have been fixed to not trigger use-after-free errors,
# this decorator can be retired.
frequently_corrupts_memory = (
    skip('frequently corrupts memory')
    if is_mac_os() and is_ci() and not is_asan()
    else (lambda f: f)
)


# ------------------------------------------------------------------------------
# Test: capture_crashes_to*

async def test_capture_crashes_to_self_decorator_works() -> None:
    class MyTask(_DownloadResourcesPlaceholderTask):
        def __init__(self) -> None:
            super().__init__(1)
        
        @capture_crashes_to_self
        def foo_did_bar(self) -> None:
            raise _CRASH
    my_task = MyTask()
    
    assert my_task.crash_reason is None
    my_task.foo_did_bar()
    assert my_task.crash_reason is not None


async def test_capture_crashes_to_self_decorator_with_custom_return_value_works() -> None:
    class MyTask(_DownloadResourcesPlaceholderTask):
        def __init__(self) -> None:
            super().__init__(1)
        
        @capture_crashes_to_self(return_if_crashed=Ellipsis)
        def calculate_foo(self) -> int:
            if False:
                return 1
            else:
                raise _CRASH
    my_task = MyTask()
    
    assert my_task.crash_reason is None
    assert Ellipsis == my_task.calculate_foo()
    assert my_task.crash_reason is not None


async def test_capture_crashes_to_bulkhead_arg_decorator_works() -> None:
    class MyTask(_DownloadResourcesPlaceholderTask):
        def __init__(self) -> None:
            super().__init__(1)
        
        @capture_crashes_to_task_arg
        def task_foo_did_change(self, task: Task) -> None:
            raise _CRASH
    my_task = MyTask()
    other_task = _DownloadResourcesPlaceholderTask(1)
    
    assert other_task.crash_reason is None
    assert my_task.crash_reason is None
    
    my_task.task_foo_did_change(other_task)
    
    assert other_task.crash_reason is not None
    assert my_task.crash_reason is None


async def test_capture_crashes_to_bulkhead_arg_decorator_with_custom_return_value_works() -> None:
    class MyTask(_DownloadResourcesPlaceholderTask):
        def __init__(self) -> None:
            super().__init__(1)
        
        @capture_crashes_to_task_arg(return_if_crashed=Ellipsis)
        def calculate_foo(self, task: Task) -> int:
            if False:
                return 1
            else:
                raise _CRASH
    my_task = MyTask()
    other_task = _DownloadResourcesPlaceholderTask(1)
    
    assert other_task.crash_reason is None
    assert my_task.crash_reason is None
    
    assert Ellipsis == my_task.calculate_foo(other_task)
    
    assert other_task.crash_reason is not None
    assert my_task.crash_reason is None


async def test_capture_crashes_to_decorator_works() -> None:
    class MyTask(_DownloadResourcesPlaceholderTask):
        def __init__(self) -> None:
            super().__init__(1)
        
        def foo_did_bar(self) -> None:
            @capture_crashes_to(self)
            def fg_task() -> None:
                raise _CRASH
            fg_call_and_wait(fg_task)
    my_task = MyTask()
    
    assert my_task.crash_reason is None
    my_task.foo_did_bar()
    assert my_task.crash_reason is not None


async def test_capture_crashes_to_decorator_with_custom_return_value_works() -> None:
    bulkhead = BulkheadCell()
    
    @capture_crashes_to(bulkhead, return_if_crashed=Ellipsis)
    def calculate_foo() -> int:
        if False:
            return 1
        else:
            raise _CRASH
    
    assert bulkhead.crash_reason is None
    assert Ellipsis == calculate_foo()
    assert bulkhead.crash_reason is not None


def test_crashes_captured_to_context_manager_works() -> None:
    bulkhead = BulkheadCell()
    assert None == bulkhead.crash_reason
    
    try:
        with crashes_captured_to(bulkhead):
            pass
    except NotImplementedError:
        pass
    else:
        raise AssertionError(
            'Expected crashes_captured_to() to raise NotImplementedError')
    
    with crashes_captured_to(bulkhead, enter_if_crashed=True):
        pass
    assert None == bulkhead.crash_reason
    
    with crashes_captured_to(bulkhead, enter_if_crashed=True):
        def foo() -> None:
            def bar() -> None:
                raise _CRASH
            bar()
        foo()
    assert _CRASH == bulkhead.crash_reason


async def test_capture_crashes_to_stderr_decorator_works() -> None:
    @capture_crashes_to_stderr
    def report_error() -> None:
        raise _CRASH
    
    with redirect_stderr(StringIO()) as captured_stderr:
        report_error()
    assert 'Exception in bulkhead:' in captured_stderr.getvalue()
    assert 'Traceback' in captured_stderr.getvalue()
    assert cli.TERMINAL_FG_RED in captured_stderr.getvalue()


async def test_capture_crashes_to_stderr_decorator_with_custom_return_value_works() -> None:
    @capture_crashes_to_stderr(return_if_crashed=Ellipsis)
    def calculate_foo() -> int:
        if False:
            return 1
        else:
            raise _CRASH
    
    with redirect_stderr(StringIO()) as captured_stderr:
        assert Ellipsis == calculate_foo()
    assert 'Exception in bulkhead:' in captured_stderr.getvalue()
    assert 'Traceback' in captured_stderr.getvalue()
    assert cli.TERMINAL_FG_RED in captured_stderr.getvalue()


async def test_given_callable_decorated_by_capture_crashes_to_star_decorator_when_run_bulkhead_call_used_then_calls_callable() -> None:
    @capture_crashes_to_stderr
    def protected_method():
        protected_method.called = True
    
    def unprotected_method():
        unprotected_method.called = True
    
    run_bulkhead_call(protected_method)
    assert True == getattr(protected_method, 'called', False)
    
    try:
        run_bulkhead_call(unprotected_method)
    except AssertionError:
        pass  # expected
    assert False == getattr(unprotected_method, 'called', False)


@skip('covered by: test_given_callable_decorated_by_capture_crashes_to_star_decorator_when_run_bulkhead_call_used_then_calls_callable')
async def test_given_not_callable_decorated_by_capture_crashes_to_star_decorator_when_run_bulkhead_call_used_then_raises() -> None:
    pass


# ------------------------------------------------------------------------------
# Test: Print Unhandled Exceptions
#
# In general:
# - Tracebacks for all unhandled exceptions should be printed to the stderr.
# - If such an exception is forwarded to the GUI, it need only be printed in
#   a yellow warning color, since it should be visible to GUI-only users.
# - If such an exception is NOT forwarded to the GUI, it should be printed in
#   a red error color, since it is INVISIBLE to GUI-only users.

async def test_given_inside_bulkhead_when_unhandled_exception_raised_then_traceback_printed_to_stderr_as_yellow_warning() -> None:
    with redirect_stderr(StringIO()) as captured_stderr:
        class MyTask(_DownloadResourcesPlaceholderTask):
            def __init__(self) -> None:
                super().__init__(1)
            
            @capture_crashes_to_self
            def protected_method(self) -> None:
                raise _CRASH
        
        MyTask().protected_method()
    assert 'Exception in bulkhead:' in captured_stderr.getvalue()
    assert 'Traceback' in captured_stderr.getvalue()
    assert cli.TERMINAL_FG_YELLOW in captured_stderr.getvalue()


async def test_given_inside_wxpython_listener_when_exception_raised_then_traceback_printed_to_stderr_as_red_error() -> None:
    frame = wx.Frame(None, title='Button Frame')
    try:
        button = wx.Button(frame, label='Button')
        def action_func(event: wx.CommandEvent):
            raise _CRASH
        bind(button, wx.EVT_BUTTON, action_func)
        
        assert is_foreground_thread()
        with redirect_stderr(StringIO()) as captured_stderr:
            click_button(button)
        assert 'Exception in wxPython listener:' in captured_stderr.getvalue()
        assert 'Traceback' in captured_stderr.getvalue()
        assert cli.TERMINAL_FG_RED in captured_stderr.getvalue()
    finally:
        frame.Destroy()


async def test_given_inside_background_thread_when_exception_raised_then_traceback_printed_to_stderr_as_red_error() -> None:
    def bg_task() -> None:
        raise _CRASH
    with redirect_stderr(StringIO()) as captured_stderr:
        # NOTE: Most creators of background threads in Crystal prefer to use
        #       bg_call_later() rather than creating a Thread directly.
        #       
        #       bg_call_later() already enforces that its callable
        #       is wrapped with @capture_crashes_to* and reports any
        #       unhandled exceptions somewhere sensible, but bare
        #       Threads don't.
        thread = threading.Thread(target=bg_task)
        thread.start()
        thread.join()
    assert 'Exception in background thread:' in captured_stderr.getvalue()
    assert 'Traceback' in captured_stderr.getvalue()
    assert cli.TERMINAL_FG_RED in captured_stderr.getvalue()


async def test_given_inside_main_thread_when_exception_raised_then_traceback_printed_to_stderr_as_red_error() -> None:
    try:
        raise _CRASH
    except Exception as e_:
        e = e_  # capture
    
    with redirect_stderr(StringIO()) as captured_stderr:
        sys.excepthook(type(e), e, e.__traceback__)
    assert 'Exception in main thread:' in captured_stderr.getvalue()
    assert 'Traceback' in captured_stderr.getvalue()
    assert cli.TERMINAL_FG_RED in captured_stderr.getvalue()


async def test_given_inside_unraisable_context_when_exception_raised_then_traceback_printed_to_stderr_as_red_error() -> None:
    def object_holder() -> None:
        class CrashesDuringFinalization:
            def __del__(self) -> None:
                raise _CRASH
        CrashesDuringFinalization()
    
    with redirect_stderr(StringIO()) as captured_stderr:
        object_holder()
    assert 'Exception in unraisable context:' in captured_stderr.getvalue()
    assert 'Traceback' in captured_stderr.getvalue()
    assert cli.TERMINAL_FG_RED in captured_stderr.getvalue()


# ------------------------------------------------------------------------------
# Test: Common Crash Locations in Task
# 
# Below, some abbreviations are used in test names:
# - DRGMT = DownloadResourceGroupMembersTask
# - DRGT = DownloadResourceGroupTask
# - DRT = DownloadResourceTask
# - T = Task

@frequently_corrupts_memory
async def test_when_T_try_get_next_task_unit_crashes_then_T_displays_as_crashed() -> None:
    with scheduler_disabled(), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                # Create DownloadResourceTask
                home_r = Resource(project, home_url)
                home_r.download(); append_deferred_top_level_tasks(project)
                (download_r_task,) = project.root_task.children
                assert isinstance(download_r_task, DownloadResourceTask)
                
                # Force DownloadResourceTask into an illegal state
                # (i.e. a container task with empty children)
                # which should crash the next call to DRG.try_get_next_task_unit()
                assert len(download_r_task.children) > 0
                download_r_task._children = []
                
                # Precondition
                assert download_r_task.crash_reason is None
                
                await step_scheduler(project, expect_done=True)
                
                # Postcondition
                assert download_r_task.crash_reason is not None


@frequently_corrupts_memory
async def test_when_DRT_child_task_did_complete_event_crashes_then_DRT_displays_as_crashed() -> None:
    with scheduler_disabled(), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                # Create DownloadResourceTask
                home_r = Resource(project, home_url)
                home_r.download(); append_deferred_top_level_tasks(project)
                (download_r_task,) = project.root_task.children
                assert isinstance(download_r_task, DownloadResourceTask)
                
                # Precondition
                assert download_r_task.crash_reason is None
                
                # Download URL
                await step_scheduler(project)
                
                # Parse links
                #
                # Patch urljoin() to simulate effect of calling urlparse('//[oops'),
                # which raises an exception in stock Python:
                # https://discuss.python.org/t/urlparse-can-sometimes-raise-an-exception-should-it/44465
                # 
                # NOTE: Overrides the fix in commit 5aaaba57076d537a4872bb3cf7270112ca497a06,
                #       reintroducing the related bug it fixed.
                with patch('crystal.task.urljoin', side_effect=ValueError('Invalid IPv6 URL')):
                    await step_scheduler(project)
                
                # Postcondition:
                # Ensure crashed in DownloadResourceTask.child_task_did_complete(),
                # when tried to resolve relative_urls from links parsed by
                # ParseResourceRevisionLinks to absolute URLs
                assert download_r_task.crash_reason is not None


@frequently_corrupts_memory
async def test_when_DRGMT_load_children_crashes_then_DRGT_displays_as_crashed() -> None:
    with scheduler_disabled(), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
        rss_feed_url = sp.get_request_url('https://xkcd.com/rss.xml')
        feed_pattern = sp.get_request_url('https://xkcd.com/*.xml')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                Resource(project, atom_feed_url)
                Resource(project, rss_feed_url)
                
                # Create DownloadResourceGroupMembersTask
                feed_g = ResourceGroup(project, '', feed_pattern, source=None)
                feed_g.download(); append_deferred_top_level_tasks(project)
                (download_rg_task,) = project.root_task.children
                assert isinstance(download_rg_task, DownloadResourceGroupTask)
                (update_rg_members_task, download_rg_members_task) = download_rg_task.children
                assert isinstance(update_rg_members_task, UpdateResourceGroupMembersTask)
                assert isinstance(download_rg_members_task, DownloadResourceGroupMembersTask)
                
                # Capture download_rg_members_task.{_pdb, _children_loaded, subtitle}
                # after exiting DownloadResourceGroupMembersTask.__init__()
                pbc_after_init = download_rg_members_task._pbc
                children_loaded_after_init = download_rg_members_task._children_loaded
                subtitle_after_init = download_rg_members_task.subtitle
                assert None == pbc_after_init
                assert False == children_loaded_after_init
                assert 'Queued' == subtitle_after_init
                
                # Precondition
                assert download_rg_members_task.crash_reason is None
                
                # Load children of DownloadResourceGroupMembersTask
                # 
                # Patch DownloadResourceGroupMembersTask.initialize_children() to reintroduce
                # a bug in DownloadResourceGroupMembersTask._load_children() that was
                # fixed in commit 44f5bd429201972d324df1287e673ddef9ffa936
                with patch.object(download_rg_members_task, 'initialize_children', side_effect=_CRASH):
                    await step_scheduler(project)
                
                # Postcondition
                assert download_rg_members_task.crash_reason is not None


@frequently_corrupts_memory
async def test_when_DRGMT_group_did_add_member_event_crashes_then_DRGT_displays_as_crashed() -> None:
    with scheduler_disabled(), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
        rss_feed_url = sp.get_request_url('https://xkcd.com/rss.xml')
        feed_pattern = sp.get_request_url('https://xkcd.com/*.xml')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                Resource(project, atom_feed_url)
                
                # Create DownloadResourceGroupMembersTask
                feed_g = ResourceGroup(project, '', feed_pattern, source=None)
                feed_g.download(); append_deferred_top_level_tasks(project)
                (download_rg_task,) = project.root_task.children
                assert isinstance(download_rg_task, DownloadResourceGroupTask)
                (update_rg_members_task, download_rg_members_task) = download_rg_task.children
                assert isinstance(update_rg_members_task, UpdateResourceGroupMembersTask)
                assert isinstance(download_rg_members_task, DownloadResourceGroupMembersTask)
                
                super_notify_did_append_child = download_rg_members_task.notify_did_append_child
                def notify_did_append_child(*args, **kwargs):
                    super_notify_did_append_child(*args, **kwargs)
                    
                    # Simulate failure of `self._pbc.total += 1`
                    # due to self._pbc being unset, due to known issue:
                    # https://github.com/davidfstr/Crystal-Web-Archiver/issues/141
                    # 
                    # TODO: Once that issue is fixed, replace the next line with:
                    #           raise _CRASH
                    raise AttributeError("'DownloadResourceGroupMembersTask' object has no attribute '_pbc'")
                
                # Precondition
                assert download_rg_members_task.crash_reason is None
                
                with patch.object(download_rg_members_task, 'notify_did_append_child', notify_did_append_child):
                    Resource(project, rss_feed_url)
                
                # Postcondition
                assert download_rg_members_task.crash_reason is not None


# ------------------------------------------------------------------------------
# Test: Common Crash Locations in TaskTreeNode
# 
# Below, some abbreviations are used in test names:
# - T = Task
# - TTN = TaskTreeNode

@skip('not yet automated: hard to automate: hard to patch')
async def test_when_TTN_task_crash_reason_did_change_crashes_at_top_level_then_crash_reason_printed_to_stderr() -> None:
    # TODO: In TaskTreeNode.task_crash_reason_did_change,
    #       crash the line: fg_call_later(fg_task, profile=False)
    #       even though that specific line is unrealistic,
    #       because there isn't a better more-realistic candidate line available
    pass


@frequently_corrupts_memory
async def test_when_TTN_task_crash_reason_did_change_crashes_in_deferred_fg_task_then_crash_reason_printed_to_stderr() -> None:
    with scheduler_disabled(), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                # Create DownloadResourceTask and TaskTreeNode
                home_r = Resource(project, home_url)
                home_r.download(); append_deferred_top_level_tasks(project)
                (download_r_task,) = project.root_task.children
                assert isinstance(download_r_task, DownloadResourceTask)
                ttn_for_task(download_r_task)
                
                # Force DownloadResourceTask into an illegal state
                # (i.e. a container task with empty children)
                # which should crash the next call to DRG.try_get_next_task_unit()
                assert len(download_r_task.children) > 0
                download_r_task._children = []
                
                # Precondition
                assert download_r_task.crash_reason is None
                
                # In TaskTreeNode.task_crash_reason_did_change,
                # crash the line: ... = self._calculate_tree_node_subtitle(task_subtitle, task_crash_reason)
                with patch('crystal.browser.tasktree.TaskTreeNode._calculate_tree_node_subtitle', side_effect=_CRASH) as crash_point:
                    with redirect_stderr(StringIO()) as captured_stderr:
                        await step_scheduler(project, expect_done=True)
                
                # Postconditions
                if True:
                    assert download_r_task.crash_reason is not None
                    assert crash_point.call_count >= 1
                    
                    assert 'Exception in bulkhead:' in captured_stderr.getvalue()
                    assert 'Traceback' in captured_stderr.getvalue()
                    assert cli.TERMINAL_FG_RED in captured_stderr.getvalue()


@frequently_corrupts_memory
async def test_when_TTN_task_did_set_children_crashes_at_top_level_then_T_displays_as_crashed() -> None:
    with scheduler_disabled(), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
        rss_feed_url = sp.get_request_url('https://xkcd.com/rss.xml')
        feed_pattern = sp.get_request_url('https://xkcd.com/*.xml')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                Resource(project, atom_feed_url)
                Resource(project, rss_feed_url)
                
                # Create DownloadResourceGroupMembersTask
                feed_g = ResourceGroup(project, '', feed_pattern, source=None)
                feed_g.download(); append_deferred_top_level_tasks(project)
                (download_rg_task,) = project.root_task.children
                assert isinstance(download_rg_task, DownloadResourceGroupTask)
                (update_rg_members_task, download_rg_members_task) = download_rg_task.children
                assert isinstance(update_rg_members_task, UpdateResourceGroupMembersTask)
                assert isinstance(download_rg_members_task, DownloadResourceGroupMembersTask)
                
                # Precondition
                assert download_rg_members_task.crash_reason is None
                
                # In TaskTreeNode.task_did_set_children,
                # crash the assertion: assert self._num_visible_children == visible_child_count
                if True:
                    super_task_did_set_children = TaskTreeNode.task_did_set_children
                    # NOTE: Need some kind of @capture_crashes_to* here to pass caller checks
                    @capture_crashes_to_stderr
                    def task_did_set_children(self, task: Task, child_count: int) -> None:
                        # Corrupt the value of self._num_visible_children
                        self._num_visible_children += 100
                        return super_task_did_set_children(self, task, child_count)
                    
                    # Load children of DownloadResourceGroupMembersTask
                    with patch.object(TaskTreeNode, 'task_did_set_children', task_did_set_children):
                        await step_scheduler(project)
                
                # Postcondition
                assert download_rg_members_task.crash_reason is not None


@frequently_corrupts_memory
async def test_when_TTN_task_did_set_children_crashes_in_deferred_fg_task_then_T_displays_as_crashed() -> None:
    with scheduler_disabled(), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
        rss_feed_url = sp.get_request_url('https://xkcd.com/rss.xml')
        feed_pattern = sp.get_request_url('https://xkcd.com/*.xml')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                Resource(project, atom_feed_url)
                Resource(project, rss_feed_url)
                
                # Create DownloadResourceGroupMembersTask
                feed_g = ResourceGroup(project, '', feed_pattern, source=None)
                feed_g.download(); append_deferred_top_level_tasks(project)
                (download_rg_task,) = project.root_task.children
                assert isinstance(download_rg_task, DownloadResourceGroupTask)
                (update_rg_members_task, download_rg_members_task) = download_rg_task.children
                assert isinstance(update_rg_members_task, UpdateResourceGroupMembersTask)
                assert isinstance(download_rg_members_task, DownloadResourceGroupMembersTask)
                
                # Precondition
                assert download_rg_members_task.crash_reason is None
                
                # Load children of DownloadResourceGroupMembersTask
                # 
                # In TaskTreeNode.task_did_set_children,
                # crash the call: self.tree_node.append_child(...)
                with patch.object(NodeView, 'append_child', side_effect=_CRASH) as crash_point:
                    await step_scheduler(project)
                    await pump_wx_events()  # force deferral by fg_call_later() to run
                
                # Postcondition
                assert crash_point.call_count >= 1
                assert download_rg_members_task.crash_reason is not None


@frequently_corrupts_memory
async def test_when_TTN_task_did_append_child_crashes_at_top_level_then_T_displays_as_crashed() -> None:
    with scheduler_disabled(), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                # Create DownloadResourceTask and TaskTreeNode
                home_r = Resource(project, home_url)
                home_r.download(); append_deferred_top_level_tasks(project)
                (download_r_task,) = project.root_task.children
                assert isinstance(download_r_task, DownloadResourceTask)
                ttn_for_task(download_r_task)
                
                # Precondition
                assert download_r_task.crash_reason is None
                
                # In TaskTreeNode.task_did_append_child,
                # crash the line: child = task.children[-1]  # lookup child
                if True:
                    super_task_did_append_child = TaskTreeNode.task_did_append_child
                    # NOTE: Need some kind of @capture_crashes_to* here to pass caller checks
                    @capture_crashes_to_stderr
                    def task_did_append_child(self, task: Task, child: Task | None) -> None:
                        # Corrupt the value of task.children
                        task._children = []
                        child = None  # force access of task.children
                        return super_task_did_append_child(self, task, child)
                    
                    # Load children of DownloadResourceGroupMembersTask
                    with patch.object(TaskTreeNode, 'task_did_append_child', task_did_append_child):
                        await step_scheduler(project)
                
                # Postcondition
                assert download_r_task.crash_reason is not None


@frequently_corrupts_memory
async def test_when_TTN_task_did_append_child_crashes_in_deferred_fg_task_then_T_displays_as_crashed() -> None:
    with scheduler_disabled(), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                # Create DownloadResourceTask and TaskTreeNode
                home_r = Resource(project, home_url)
                home_r.download(); append_deferred_top_level_tasks(project)
                (download_r_task,) = project.root_task.children
                assert isinstance(download_r_task, DownloadResourceTask)
                ttn_for_task(download_r_task)
                
                # Precondition
                assert download_r_task.crash_reason is None
                
                # In TaskTreeNode.task_did_append_child,
                # Crash the line: self.tree_node.append_child(child_ttnode.tree_node)
                with patch.object(NodeView, 'append_child', side_effect=_CRASH) as crash_point:
                    await step_scheduler(project)
                    await pump_wx_events()  # force deferral by fg_call_later() to run
                
                # Postcondition
                assert crash_point.call_count >= 1
                assert download_r_task.crash_reason is not None


@frequently_corrupts_memory
async def test_when_TTN_task_child_did_complete_crashes_at_top_level_then_T_displays_as_crashed() -> None:
    with scheduler_disabled(), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                # Create DownloadResourceTask and TaskTreeNode
                home_r = Resource(project, home_url)
                home_r.download(); append_deferred_top_level_tasks(project)
                (download_r_task,) = project.root_task.children
                assert isinstance(download_r_task, DownloadResourceTask)
                ttn_for_task(download_r_task)
                
                # Precondition
                assert download_r_task.crash_reason is None
                
                # In TaskTreeNode.task_child_did_complete,
                # crash the line: with self._complete_events_ignored():
                with patch.object(TaskTreeNode, '_complete_events_ignored', side_effect=_CRASH) as crash_point:
                    await step_scheduler(project)
                    await pump_wx_events()  # force deferral by fg_call_later() to run
                
                # Postcondition
                assert crash_point.call_count >= 1
                assert download_r_task.crash_reason is not None


@frequently_corrupts_memory
async def test_when_TTN_task_child_did_complete_crashes_in_deferred_fg_task_then_T_displays_as_crashed() -> None:
    with scheduler_disabled(), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                # Create DownloadResourceTask and TaskTreeNode
                home_r = Resource(project, home_url)
                home_r.download(); append_deferred_top_level_tasks(project)
                (download_r_task,) = project.root_task.children
                assert isinstance(download_r_task, DownloadResourceTask)
                ttn_for_task(download_r_task)
                
                # Precondition
                assert download_r_task.crash_reason is None
                
                # In TaskTreeNode.task_child_did_complete,
                # crash the line: self.tree_node.children = new_children
                with patch.object(TaskTreeNode, '_MAX_LEADING_COMPLETE_CHILDREN', 0), \
                        patch.object(NodeView, 'set_children', side_effect=_CRASH) as crash_point:
                    await step_scheduler(project)
                    await pump_wx_events()  # force deferral by fg_call_later() to run
                
                # Postcondition
                assert crash_point.call_count >= 1
                assert download_r_task.crash_reason is not None


@skip('not yet automated')
async def test_when_TTN_task_did_clear_children_crashes_at_top_level_then_T_displays_as_crashed() -> None:
    # TODO: Crash the line: if self.task.scheduling_style == SCHEDULING_STYLE_SEQUENTIAL: raise NotImplementedError()
    pass


@skip('not yet automated')
async def test_when_TTN_task_did_clear_children_crashes_in_deferred_fg_task_then_T_displays_as_crashed() -> None:
    # TODO: Crash the line: self.tree_node.children = [... if i not in child_indexes]
    pass


# ------------------------------------------------------------------------------
# Test: Common Crash Locations in entitytree.Node
# 
# Below, some abbreviations are used in test names:
# - ET = EntityTree
# - RGN = ResourceGroupNode
# - RN = _ResourceNode

@skip('not yet automated')
async def test_when_RN_on_expanded_crashes_at_top_level_then_children_replaced_with_error_node() -> None:
    # TODO: In _ResourceNode.on_expanded,
    #       crash the line: self.download_future = self.resource.download()
    pass


@skip('not yet automated')
async def test_when_RN_on_expanded_crashes_while_updating_children_then_children_replaced_with_error_node() -> None:
    # TODO: In _ResourceNode.update_children,
    #       crash the line: children.append(RootResourceNode(...))
    pass


@frequently_corrupts_memory
async def test_when_RGN_on_expanded_crashes_while_loading_urls_then_children_replaced_with_error_node() -> None:
    with scheduler_disabled(), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
        rss_feed_url = sp.get_request_url('https://xkcd.com/rss.xml')
        feed_pattern = sp.get_request_url('https://xkcd.com/*.xml')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                Resource(project, atom_feed_url)
                Resource(project, rss_feed_url)
                
                # Create ResourceGroupNode
                feed_g = ResourceGroup(project, '', feed_pattern, source=None)
                root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                feed_ti = root_ti.find_child(feed_pattern)
                
                # In ResourceGroupNode.on_expanded,
                # crash the line: self.resource_group.project.load_urls()
                with patch.object(project, 'load_urls', side_effect=_CRASH):
                    feed_ti.Expand()
                    await wait_for(first_child_of_tree_item_is_not_loading_condition(feed_ti))
                
                # Postconditions
                (error_ti,) = feed_ti.Children
                assert _ErrorNode.CRASH_TITLE == error_ti.Text
                assert _ErrorNode.CRASH_TEXT_COLOR == error_ti.TextColour
                assert True == error_ti.Bold


@frequently_corrupts_memory
async def test_when_RGN_on_expanded_crashes_while_updating_children_then_children_replaced_with_error_node() -> None:
    with scheduler_disabled(), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
        rss_feed_url = sp.get_request_url('https://xkcd.com/rss.xml')
        feed_pattern = sp.get_request_url('https://xkcd.com/*.xml')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                Resource(project, atom_feed_url)
                Resource(project, rss_feed_url)
                
                # Create ResourceGroupNode
                feed_g = ResourceGroup(project, '', feed_pattern, source=None)
                root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                feed_ti = root_ti.find_child(feed_pattern)
                
                # In ResourceGroupNode.update_children,
                # crash the line: children_rs.append(NormalResourceNode(...))
                with patch('crystal.browser.entitytree.NormalResourceNode', side_effect=_CRASH):
                    feed_ti.Expand()
                    await wait_for(first_child_of_tree_item_is_not_loading_condition(feed_ti))
                
                # Postconditions
                (error_ti,) = feed_ti.Children
                assert _ErrorNode.CRASH_TITLE == error_ti.Text
                assert _ErrorNode.CRASH_TEXT_COLOR == error_ti.TextColour
                assert True == error_ti.Bold


@frequently_corrupts_memory
async def test_when_RGN_update_children_crashes_during_ET_resource_did_instantiate_then_RGN_children_replaced_with_error_node() -> None:
    with scheduler_disabled(), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
        rss_feed_url = sp.get_request_url('https://xkcd.com/rss.xml')
        feed_pattern = sp.get_request_url('https://xkcd.com/*.xml')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                Resource(project, atom_feed_url)
                
                # Create ResourceGroupNode
                feed_g = ResourceGroup(project, '', feed_pattern, source=None)
                root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                feed_ti = root_ti.find_child(feed_pattern)
                
                # Expand ResourceGroupNode
                feed_ti.Expand()
                await wait_for(first_child_of_tree_item_is_not_loading_condition(feed_ti))
                
                super_update_children = ResourceGroupNode.update_children
                # NOTE: Need some kind of @capture_crashes_to* here to pass caller checks
                @capture_crashes_to_stderr
                def update_children(self, *args, **kwargs) -> None:
                    update_children.called = True  # type: ignore[attr-defined]
                    return super_update_children(self, *args, **kwargs)
                update_children.called = False  # type: ignore[attr-defined]
                
                # In ResourceGroupNode.update_children,
                # crash the line: children_rs.append(NormalResourceNode(...))
                with patch('crystal.browser.entitytree.NormalResourceNode', side_effect=_CRASH), \
                        patch.object(ResourceGroupNode, 'update_children', update_children):
                    Resource(project, rss_feed_url)
                    await wait_for(lambda: update_children.called or None)  # type: ignore[attr-defined]
                
                # Postconditions
                (error_ti,) = feed_ti.Children
                assert _ErrorNode.CRASH_TITLE == error_ti.Text
                assert _ErrorNode.CRASH_TEXT_COLOR == error_ti.TextColour
                assert True == error_ti.Bold


# ------------------------------------------------------------------------------
# Test: Common Crash Locations in EntityTree

# NOTE: This has not been observed to be a *common* crash location.
#       It is, however, the only test currently covering what happens what
#       a crash happens at the level of the EntityTree itself (as opposed
#       to inside one of its nodes).
@frequently_corrupts_memory
async def test_when_ET_root_resource_did_instantiate_crashes_then_updating_entity_tree_crashed_task_appears() -> None:
    with scheduler_disabled(), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            rmw = RealMainWindow._last_created
            assert rmw is not None
            
            with clear_top_level_tasks_on_exit(project):
                # Preconditions
                et_root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                () = et_root_ti.Children
                
                with patch('crystal.browser.entitytree.RootResourceNode', side_effect=_CRASH):
                    RootResource(project, 'Home', Resource(project, home_url))
                
                # Postconditions
                if True:
                    assert rmw.entity_tree.crash_reason is not None
                    
                    et_root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                    () = et_root_ti.Children
                    
                    project.root_task.append_deferred_top_level_tasks()
                    (updating_et_crashed_task,) = project.root_task.children
                    assert isinstance(updating_et_crashed_task, CrashedTask)
                
                # test_given_updating_entity_tree_crashed_task_at_top_level_when_right_click_task_then_menu_appears_with_enabled_refresh_menuitem
                root_ti = TreeItem.GetRootItem(mw.task_tree)
                (scheduler_crashed_ti,) = root_ti.Children
                def show_popup(menu: wx.Menu) -> None:
                    (refresh_menuitem,) = (
                        mi for mi in menu.MenuItems
                        if mi.ItemLabelText == 'Refresh'
                    )
                    assert refresh_menuitem.Enabled
                    
                    # test_when_click_refresh_menuitem_for_updating_entity_tree_crashed_task_then_recreates_unexpanded_top_level_entities_in_entity_tree_and_task_is_removed
                    select_menuitem_now(menu, refresh_menuitem.Id)
                    if True:
                        assert rmw.entity_tree.crash_reason is None
                        
                        et_root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                        (home_ti,) = et_root_ti.Children
                        
                        # RT -- notices all finished children; clears them
                        step_scheduler_now(project, expect_done=True)
                        
                        () = project.root_task.children
                await scheduler_crashed_ti.right_click_showing_popup_menu(show_popup)


# ------------------------------------------------------------------------------
# Test: Crashes Cascade to Ancestors
# 
# Below, some abbreviations are used in test names:
# - DRBT = DownloadResourceBodyTask
# - DRGMT = DownloadResourceGroupMembersTask
# - DRGT = DownloadResourceGroupTask
# - DRT = DownloadResourceTask
# - PRRL = ParseResourceRevisionLinks
# - RT = RootTask
# - URGMT = UpdateResourceGroupMembersTask

@skip('not yet automated: hard to crash DRT (or any leaf task) in a realistic way')
async def test_when_DRBT_child_of_DRT_crashes_then_DRT_displays_as_crashed() -> None:
    pass


@skip('not yet automated: hard to crash PRRL (or any leaf task) in a realistic way')
async def test_when_PRRL_child_of_DRT_crashes_then_DRT_displays_as_crashed() -> None:
    pass


@frequently_corrupts_memory
async def test_when_DRT_child_of_DRT_crashes_then_parent_DRT_displays_as_crashed() -> None:
    assert (
        task.SCHEDULING_STYLE_SEQUENTIAL == 
        DownloadResourceTask.scheduling_style
    )
    
    with scheduler_disabled(), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                # Create DownloadResourceTask
                home_r = Resource(project, home_url)
                home_r.download(); append_deferred_top_level_tasks(project)
                (download_r_task,) = project.root_task.children
                assert isinstance(download_r_task, DownloadResourceTask)
                
                # Precondition
                assert download_r_task.crash_reason is None
                
                # RT > DRT > DownloadResourceBodyTask
                def check() -> None:
                    assert not download_r_task.children[0].complete
                await step_scheduler(project, after_get=check)
                assert download_r_task.children[0].complete
                
                # RT > DRT > ParseResourceRevisionLinks
                def check() -> None:
                    assert not download_r_task.children[1].complete
                await step_scheduler(project, after_get=check)
                assert download_r_task.children[1].complete
                assert len(download_r_task.children) >= 3
                
                # RT > DRT > DRT[0] > DownloadResourceBodyTask
                def check() -> None:
                    assert not download_r_task.children[2].children[0].complete
                await step_scheduler(project, after_get=check)
                assert download_r_task.children[2].children[0].complete
                
                # RT > DRT > DRT[0] > ParseResourceRevisionLinks
                await _step_scheduler_with_unit_from_PRRL_and_crash_task(
                    download_r_task.children[2].children[1],
                    home_r)
                
                await step_scheduler(project, expect_done=True)
                
                # Postconditions:
                # 1. Ensure crashed in DownloadResourceTask.child_task_did_complete(),
                #    when tried to resolve relative_urls from links parsed by
                #    ParseResourceRevisionLinks to absolute URLs
                # 2. Ensure crash cascades to parent of inner DownloadResourceTask
                # 3. Ensure crash cascade stops just before reaching the RootTask
                assert download_r_task.children[2].crash_reason is not None
                assert download_r_task.crash_reason is not None
                assert project.root_task.crash_reason is None
                
                root_ti = TreeItem.GetRootItem(mw.task_tree)
                (download_r_ti,) = root_ti.Children
                
                # test_when_hover_mouse_over_crashed_task_then_tooltip_with_user_facing_traceback_appears
                tooltip = download_r_ti.Tooltip()
                assert tooltip is not None
                assert 'ValueError: Invalid IPv6 URL' in tooltip, f'Tooltip was: {tooltip}'
                assert (
                    'at crystal/task.py:' in tooltip or
                    'at crystal/task.pyc:' in tooltip or
                    r'at crystal\task.pyc:' in tooltip
                ), f'Tooltip was: {tooltip}'
                
                # test_given_crashed_task_not_at_top_level_when_right_click_task_then_menu_appears_with_disabled_dismiss_menuitem
                download_r_ti.Expand()
                (download_rb_ti, prrl_ti, download_r1_ti, *_) = download_r_ti.Children
                assert download_r1_ti.Text.endswith(TaskTreeNode._CRASH_SUBTITLE)
                def show_popup(menu: wx.Menu) -> None:
                    (dismiss_menuitem,) = (
                        mi for mi in menu.MenuItems
                        if mi.ItemLabelText == 'Dismiss'
                    )
                    assert not dismiss_menuitem.Enabled
                await download_r1_ti.right_click_showing_popup_menu(show_popup)
                
                # test_given_regular_crashed_task_at_top_level_when_right_click_task_then_menu_appears_with_enabled_dismiss_menuitem
                def show_popup(menu: wx.Menu) -> None:
                    (dismiss_menuitem,) = (
                        mi for mi in menu.MenuItems
                        if mi.ItemLabelText == 'Dismiss'
                    )
                    assert dismiss_menuitem.Enabled
                    
                    # test_when_click_dismiss_menuitem_for_regular_top_level_crashed_task_then_task_is_removed
                    if True:
                        select_menuitem_now(menu, dismiss_menuitem.Id)
                        
                        # RT -- notices all finished children; clears them
                        step_scheduler_now(project, expect_done=True)
                        
                        () = root_ti.Children
                await download_r_ti.right_click_showing_popup_menu(show_popup)


@frequently_corrupts_memory
async def test_when_DRT_child_of_DRGMT_crashes_then_DRGMT_displays_as_crashed() -> None:
    assert (
        task.SCHEDULING_STYLE_SEQUENTIAL == 
        DownloadResourceGroupMembersTask.scheduling_style
    )
    skipTest('covered by: test_when_DRT_child_of_DRT_crashes_then_parent_DRT_displays_as_crashed')


@frequently_corrupts_memory
async def test_when_URGMT_child_of_DRGT_crashes_then_DRGT_displays_as_crashed_after_DRGMT_child_completes() -> None:
    with scheduler_disabled(), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
        rss_feed_url = sp.get_request_url('https://xkcd.com/rss.xml')
        feed_pattern = sp.get_request_url('https://xkcd.com/*.xml')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                home_r = Resource(project, home_url)
                Resource(project, atom_feed_url)
                Resource(project, rss_feed_url)
                
                # Create UpdateResourceGroupMembersTask
                home_rr = RootResource(project, 'Home', home_r)
                feed_g = ResourceGroup(project, '', feed_pattern, source=home_rr)
                feed_g.download(); append_deferred_top_level_tasks(project)
                (download_rg_task,) = project.root_task.children
                assert isinstance(download_rg_task, DownloadResourceGroupTask)
                (update_rg_members_task, download_rg_members_task) = download_rg_task.children
                assert isinstance(update_rg_members_task, UpdateResourceGroupMembersTask)
                assert isinstance(download_rg_members_task, DownloadResourceGroupMembersTask)
                
                # Preconditions
                assert download_rg_task.crash_reason is None
                assert update_rg_members_task.crash_reason is None
                
                # RT > DRGT > URGMT > DRT > DownloadResourceBodyTask
                def check() -> None:
                    assert not update_rg_members_task.children[0].children[0].complete
                await step_scheduler(project, after_get=check)
                assert update_rg_members_task.children[0].children[0].complete
                
                # RT > DRGT > DRGMT @ _load_children_and_update_completed_status
                def check() -> None:
                    assert len(download_rg_members_task.children) == 0
                await step_scheduler(project, after_get=check)
                assert len(download_rg_members_task.children) == 2
                
                # RT > DRGT > URGMT > DRT > ParseResourceRevisionLinks
                await _step_scheduler_with_unit_from_PRRL_and_crash_task(
                    update_rg_members_task.children[0].children[1],
                    home_r)
                
                # RT > DRGT > DRGMT > ... (doesn't matter)
                await step_scheduler(project)
                
                # 1. RT > DRGT > URGMT -- notices crashed child
                # 2. RT > DRGT > DRGMT > ... (doesn't matter)
                await step_scheduler(project)
                
                # Postconditions #1:
                # 1. Ensure crashed in DownloadResourceTask.child_task_did_complete(),
                #    when tried to resolve relative_urls from links parsed by
                #    ParseResourceRevisionLinks to absolute URLs
                # 2. Ensure crash cascades to URGMT
                # 3. Ensure crash does NOT cascade to DRGT yet,
                #    because DRGMT is not complete
                # 4. Ensure crash does not cascade to the RootTask
                assert update_rg_members_task.children[0].crash_reason is not None
                assert update_rg_members_task.crash_reason is not None
                assert download_rg_task.crash_reason is None
                assert project.root_task.crash_reason is None
                
                # Wait for DRGMT to complete
                while not download_rg_members_task.complete:
                    assert download_rg_members_task.crash_reason is None
                    
                    # RT > DRGT > DRGMT > ... (doesn't matter)
                    await step_scheduler(project)
                
                # RT > DRGT -- notices crashed URGMT child
                await step_scheduler(project, expect_done=True)
                
                # Postconditions #2
                # 1. Ensure crash from URGMT cascades to DRGT after DRGMT completes
                # 2. Ensure crash does not cascade to the RootTask
                assert download_rg_task.crash_reason is not None
                assert project.root_task.crash_reason is None
                
                root_ti = TreeItem.GetRootItem(mw.task_tree)
                (download_rg_ti,) = root_ti.Children
                
                # test_when_hover_mouse_over_crashed_task_then_tooltip_with_user_facing_traceback_appears
                tooltip = download_rg_ti.Tooltip()
                assert tooltip is not None
                assert 'ValueError: Invalid IPv6 URL' in tooltip, f'Tooltip was: {tooltip}'
                assert (
                    'at crystal/task.py:' in tooltip or
                    'at crystal/task.pyc:' in tooltip or
                    r'at crystal\task.pyc:' in tooltip
                ), f'Tooltip was: {tooltip}'


@frequently_corrupts_memory
async def test_when_DRGMT_child_of_DRGT_crashes_then_DRGT_displays_as_crashed_after_URGMT_child_completes() -> None:
    with scheduler_disabled(), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
        rss_feed_url = sp.get_request_url('https://xkcd.com/rss.xml')
        feed_pattern = sp.get_request_url('https://xkcd.com/*.xml')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                home_r = Resource(project, home_url)
                Resource(project, atom_feed_url)
                Resource(project, rss_feed_url)
                
                # Create DownloadResourceGroupMembersTask
                home_rr = RootResource(project, 'Home', home_r)
                feed_g = ResourceGroup(project, '', feed_pattern, source=home_rr)
                feed_g.download(); append_deferred_top_level_tasks(project)
                (download_rg_task,) = project.root_task.children
                assert isinstance(download_rg_task, DownloadResourceGroupTask)
                (update_rg_members_task, download_rg_members_task) = download_rg_task.children
                assert isinstance(update_rg_members_task, UpdateResourceGroupMembersTask)
                assert isinstance(download_rg_members_task, DownloadResourceGroupMembersTask)
                
                # Preconditions
                assert download_rg_task.crash_reason is None
                assert download_rg_members_task.crash_reason is None
                
                # RT > DRGT > URGMT > ... (doesn't matter)
                def check() -> None:
                    assert not update_rg_members_task.children[0].children[0].complete
                await step_scheduler(project, after_get=check)
                assert update_rg_members_task.children[0].children[0].complete
                
                # RT > DRGT > DRGMT @ _load_children_and_update_completed_status
                def check() -> None:
                    assert len(download_rg_members_task.children) == 0
                await step_scheduler(project, after_get=check)
                assert len(download_rg_members_task.children) == 2
                
                # RT > DRGT > URGMT > ... (doesn't matter)
                def check() -> None:
                    assert not update_rg_members_task.children[0].children[1].complete
                await step_scheduler(project, after_get=check)
                assert update_rg_members_task.children[0].children[1].complete
                
                # RT > DRGT > DRGMT > DRT > DownloadResourceBodyTask
                def check() -> None:
                    assert not download_rg_members_task.children[0].children[0].complete
                await step_scheduler(project, after_get=check)
                assert download_rg_members_task.children[0].children[0].complete
                
                # RT > DRGT > URGMT > ... (doesn't matter)
                def check() -> None:
                    assert not update_rg_members_task.children[0].children[2].children[0].complete
                await step_scheduler(project, after_get=check)
                assert update_rg_members_task.children[0].children[2].children[0].complete
                
                # RT > DRGT > DRGMT > DRT > ParseResourceRevisionLinks
                await _step_scheduler_with_unit_from_PRRL_and_crash_task(
                    download_rg_members_task.children[0].children[1],
                    home_r)
                
                # RT > DRGT > URGMT > ... (doesn't matter)
                await step_scheduler(project)
                
                # 1. RT > DRGT > DRGMT -- notices crashed child
                # 2. RT > DRGT > URGMT > ... (doesn't matter)
                await step_scheduler(project)
                
                # Postconditions #1:
                # 1. Ensure crashed in DownloadResourceTask.child_task_did_complete(),
                #    when tried to resolve relative_urls from links parsed by
                #    ParseResourceRevisionLinks to absolute URLs
                # 2. Ensure crash cascades to DRGMT
                # 3. Ensure crash does NOT cascade to DRGT yet,
                #    because URGMT is not complete
                # 4. Ensure crash does not cascade to the RootTask
                assert download_rg_members_task.children[0].crash_reason is not None
                assert download_rg_members_task.crash_reason is not None
                assert download_rg_task.crash_reason is None
                assert project.root_task.crash_reason is None
                
                # Wait for URGMT to complete
                while not update_rg_members_task.complete:
                    assert update_rg_members_task.crash_reason is None
                    
                    # RT > DRGT > URGMT > ... (doesn't matter)
                    await step_scheduler(project)
                
                # RT > DRGT -- notices crashed DRGMT child
                await step_scheduler(project, expect_done=True)
                
                # Postconditions #2
                # 1. Ensure crash from DRGMT cascades to DRGT after URGMT completes
                # 2. Ensure crash does not cascade to the RootTask
                assert download_rg_task.crash_reason is not None
                assert project.root_task.crash_reason is None
                
                root_ti = TreeItem.GetRootItem(mw.task_tree)
                (download_rg_ti,) = root_ti.Children
                
                # test_when_hover_mouse_over_crashed_task_then_tooltip_with_user_facing_traceback_appears
                tooltip = download_rg_ti.Tooltip()
                assert tooltip is not None
                assert 'ValueError: Invalid IPv6 URL' in tooltip, f'Tooltip was: {tooltip}'
                assert (
                    'at crystal/task.py:' in tooltip or
                    'at crystal/task.pyc:' in tooltip or
                    r'at crystal\task.pyc:' in tooltip
                ), f'Tooltip was: {tooltip}'


# ------------------------------------------------------------------------------
# Test: Crashes at the Top-Level and Root

@skip('covered by: test_when_DRT_child_of_DRT_crashes_then_parent_DRT_displays_as_crashed, test_when_URGMT_child_of_DRGT_crashes_then_DRGT_displays_as_crashed_after_DRGMT_child_completes, test_when_DRGMT_child_of_DRGT_crashes_then_DRGT_displays_as_crashed_after_URGMT_child_completes')
async def test_when_child_of_RT_crashes_then_RT_does_NOT_display_as_crashed() -> None:
    pass


@frequently_corrupts_memory
async def test_when_RT_try_get_next_task_unit_crashes_then_RT_marked_as_crashed() -> None:
    with scheduler_disabled(), \
            served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
        rss_feed_url = sp.get_request_url('https://xkcd.com/rss.xml')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                root_task = project.root_task
                
                # Create DownloadResourceTask #1 in RootTask, pre-appended
                home_r = Resource(project, home_url)
                home_r.download(); append_deferred_top_level_tasks(project)
                
                # Create DownloadResourceTask #2 in RootTask, un-appended
                atom_feed_r = Resource(project, atom_feed_url)
                atom_feed_r.download()
                
                # Preconditions
                assert root_task.crash_reason is None
                (download_r_task1,) = project.root_task.children
                assert isinstance(download_r_task1, DownloadResourceTask)
                
                # RT @ try_get_next_task_unit
                # 
                # In RootTask.try_get_next_task_unit,
                # in RootTask.append_deferred_top_level_tasks,
                # crash the line: super().append_child(...)
                super_append_child = Task.append_child
                def append_child(self, child: Task, *args, **kwargs) -> None:
                    if isinstance(child, DownloadResourceTask):
                        # Simulate crash when appending a DownloadResourceTask
                        append_child.call_count += 1  # type: ignore[attr-defined]
                        raise _CRASH
                    else:
                        return super_append_child(self, child, *args, **kwargs)
                append_child.call_count = 0  # type: ignore[attr-defined]
                with patch.object(Task, 'append_child', append_child):
                    await step_scheduler(project, expect_done=True)
                    assert append_child.call_count >= 1  # type: ignore[attr-defined]
                
                # Postconditions
                assert root_task.crash_reason is not None
                (download_r_task1, scheduler_crashed_task) = project.root_task.children
                assert isinstance(download_r_task1, DownloadResourceTask)
                assert isinstance(scheduler_crashed_task, CrashedTask)
                for child in root_task.children:
                    if not isinstance(child, CrashedTask) and not child.complete:
                        assert child.subtitle in ['Scheduler crashed', 'Complete'], \
                            f'Top-level task has unexpected subtitle: {child.subtitle}'
                
                root_ti = TreeItem.GetRootItem(mw.task_tree)
                (download_r_ti, scheduler_crashed_ti) = root_ti.Children
                
                # test_when_hover_mouse_over_crashed_task_then_tooltip_with_user_facing_traceback_appears
                tooltip = scheduler_crashed_ti.Tooltip()
                assert tooltip is not None
                assert 'ValueError: Simulated crash' in tooltip, f'Tooltip was: {tooltip}'
                assert (
                    'at crystal/task.py:' in tooltip or
                    'at crystal/task.pyc:' in tooltip or
                    r'at crystal\task.pyc:' in tooltip
                ), f'Tooltip was: {tooltip}'
                assert 'in append_deferred_top_level_tasks' in tooltip, f'Tooltip was: {tooltip}'
                
                # test_given_scheduler_crashed_task_at_top_level_when_right_click_task_then_menu_appears_with_enabled_dismiss_all_menuitem
                def show_popup(menu: wx.Menu) -> None:
                    (dismiss_all_menuitem,) = (
                        mi for mi in menu.MenuItems
                        if mi.ItemLabelText == 'Dismiss All'
                    )
                    assert dismiss_all_menuitem.Enabled
                    
                    # test_when_click_dismiss_all_menuitem_for_scheduler_crashed_task_then_all_top_level_tasks_are_removed
                    select_menuitem_now(menu, dismiss_all_menuitem.Id)
                    () = root_ti.Children
                await scheduler_crashed_ti.right_click_showing_popup_menu(show_popup)
                
                # ...and new top level tasks can be added that will run
                if True:
                    # Create DownloadResourceTask #3 in RootTask, pre-appended
                    rss_feed_r = Resource(project, rss_feed_url)
                    rss_feed_r.download(); append_deferred_top_level_tasks(project)
                    
                    # Preconditions
                    assert root_task.crash_reason is None
                    (download_r_task2,) = project.root_task.children
                    assert isinstance(download_r_task2, DownloadResourceTask)
                    
                    # Ensure task runs (at least one step)
                    await step_scheduler(project)


@skip('covered by: test_when_RT_try_get_next_task_unit_crashes_then_RT_marked_as_crashed')
async def test_when_RT_marked_as_crashed_then_scheduler_crashed_task_appears_and_all_other_tasks_marked_with_scheduler_crashed_subtitle() -> None:
    pass


@frequently_corrupts_memory
async def test_when_scheduler_thread_event_loop_crashes_then_RT_marked_as_crashed_and_scheduler_crashed_task_appears() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            root_task = project.root_task
            
            # Crash the scheduler event loop, for project
            super_scheduler_sleep = crystal.task.scheduler_sleep
            def scheduler_sleep(*args, **kwargs) -> None:
                f = sys._getframe(1)
                root_task_of_caller = f.f_locals['root_task']
                if root_task_of_caller == root_task and not scheduler_sleep.called:  # type: ignore[attr-defined]
                    scheduler_sleep.called = True  # type: ignore[attr-defined]
                    raise _CRASH
                else:
                    super_scheduler_sleep(*args, **kwargs)
            scheduler_sleep.called = False  # type: ignore[attr-defined]
            with patch('crystal.task.scheduler_sleep', scheduler_sleep):
                # Create DownloadResourceTask
                home_r = Resource(project, home_url)
                home_r.download()
                
                first_task_title = first_task_title_progression(mw.task_tree)
                def progression_func():
                    if root_task.crash_reason is not None:
                        # Stop
                        if (len(root_task.children) >= 1 and 
                                isinstance(root_task.children[-1], CrashedTask)):
                            return None
                        else:
                            return '(crashed; waiting for CrashedTask to appear)'
                    else:
                        # Keep waiting while first task title is changing
                        title = first_task_title()
                        if title is None:
                            return '(waiting for first task to start)'
                        else:
                            return title
                await wait_while(
                    progression_func,
                    progress_timeout=MAX_DOWNLOAD_DURATION_PER_ITEM)
            
            # Postconditions
            assert scheduler_sleep.called  # type: ignore[attr-defined]
            assert root_task.crash_reason is not None
            (*_, scheduler_crashed_task) = root_task.children
            assert isinstance(scheduler_crashed_task, CrashedTask)
            for child in root_task.children:
                if isinstance(child, CrashedTask):
                    continue
                assert child.subtitle in ['Scheduler crashed', 'Complete'], \
                    f'Top-level task has unexpected subtitle: {child.subtitle}'
            
            root_ti = TreeItem.GetRootItem(mw.task_tree)
            (*_, scheduler_crashed_ti) = root_ti.Children
            
            # test_when_hover_mouse_over_crashed_task_then_tooltip_with_user_facing_traceback_appears
            tooltip = scheduler_crashed_ti.Tooltip()
            assert tooltip is not None
            assert 'ValueError: Simulated crash' in tooltip, f'Tooltip was: {tooltip}'
            assert (
                'at crystal/task.py:' in tooltip or
                'at crystal/task.pyc:' in tooltip or
                r'at crystal\task.pyc:' in tooltip
            ), f'Tooltip was: {tooltip}'
            assert 'in bg_daemon_task' in tooltip, f'Tooltip was: {tooltip}'
            
            # Dismiss scheduler crash, clearing the task tree
            scheduler_crashed_task.dismiss()
            assert len(root_task.children) == 0


# ------------------------------------------------------------------------------
# Test: Dismissing Crashes

@skip('covered by: test_when_DRT_child_of_DRT_crashes_then_parent_DRT_displays_as_crashed')
async def test_given_regular_crashed_task_at_top_level_when_right_click_task_then_menu_appears_with_enabled_dismiss_menuitem() -> None:
    pass


@skip('covered by: test_when_DRT_child_of_DRT_crashes_then_parent_DRT_displays_as_crashed')
async def test_when_click_dismiss_menuitem_for_regular_top_level_crashed_task_then_task_is_removed() -> None:
    pass


@skip('covered by: test_when_RT_try_get_next_task_unit_crashes_then_RT_marked_as_crashed')
async def test_given_scheduler_crashed_task_at_top_level_when_right_click_task_then_menu_appears_with_enabled_dismiss_all_menuitem() -> None:
    pass


@skip('covered by: test_when_RT_try_get_next_task_unit_crashes_then_RT_marked_as_crashed')
async def test_when_click_dismiss_all_menuitem_for_scheduler_crashed_task_then_all_top_level_tasks_are_removed() -> None:
    pass


@skip('covered by: test_when_ET_root_resource_did_instantiate_crashes_then_updating_entity_tree_crashed_task_appears')
async def test_given_updating_entity_tree_crashed_task_at_top_level_when_right_click_task_then_menu_appears_with_enabled_refresh_menuitem() -> None:
    pass


@skip('covered by: test_when_ET_root_resource_did_instantiate_crashes_then_updating_entity_tree_crashed_task_appears')
async def test_when_click_refresh_menuitem_for_updating_entity_tree_crashed_task_then_recreates_unexpanded_top_level_entities_in_entity_tree_and_task_is_removed() -> None:
    pass


# TODO: Make this menuitem *enabled* iff the top-level ancestor task is crashed (very likely),
#       and make dismissing it have the same effect as dismissing the ancestor top-level task
@skip('covered by: test_when_DRT_child_of_DRT_crashes_then_parent_DRT_displays_as_crashed')
async def test_given_crashed_task_not_at_top_level_when_right_click_task_then_menu_appears_with_disabled_dismiss_menuitem() -> None:
    pass


# ------------------------------------------------------------------------------
# Test: Crash Node Tooltips

_test_names = [
    'test_when_DRT_child_of_DRT_crashes_then_parent_DRT_displays_as_crashed',
    'test_when_URGMT_child_of_DRGT_crashes_then_DRGT_displays_as_crashed_after_DRGMT_child_completes',
    'test_when_DRGMT_child_of_DRGT_crashes_then_DRGT_displays_as_crashed_after_URGMT_child_completes',
    'test_when_RT_try_get_next_task_unit_crashes_then_RT_marked_as_crashed',
    'test_when_scheduler_thread_event_loop_crashes_then_RT_marked_as_crashed_and_scheduler_crashed_task_appears',
]
@skip(f'covered by: {", ".join(_test_names)}')
async def test_when_hover_mouse_over_crashed_task_then_tooltip_with_user_facing_traceback_appears() -> None:
    pass


# ------------------------------------------------------------------------------
# Utility

async def _step_scheduler_with_unit_from_PRRL_and_crash_task(prrl_task: Task, home_r: Resource) -> None:
    """
    Performs one unit of work from the scheduler,
    where the unit of work comes from a ParseResourceRevisionLinks (PRRL) task,
    and simulate a crash when running the unit of work.
    """
    project = home_r.project
    
    assert isinstance(prrl_task, ParseResourceRevisionLinks)
    
    _PRRL_call_result__1_link = (
        [create_external_link('/', type_title='Simulated Link', title=None, embedded=True)],
        [home_r]
    )  # type: Tuple[List[Link], List[Resource]]
    
    assert not prrl_task.complete
    # 1. Force ParseResourceRevisionLinks to return 1 embedded link.
    # 2. Patch urljoin() to simulate effect of calling urlparse('//[oops'),
    #    which raises an exception in stock Python:
    #    https://discuss.python.org/t/urlparse-can-sometimes-raise-an-exception-should-it/44465
    #    
    #    NOTE: Overrides the fix in commit 5aaaba57076d537a4872bb3cf7270112ca497a06,
    #          reintroducing the related bug it fixed.
    with patch.object(prrl_task, '__call__', return_value=_PRRL_call_result__1_link) as mock_call, \
            patch('crystal.task.urljoin', side_effect=ValueError('Invalid IPv6 URL')):
        await step_scheduler(project)
        assert mock_call.call_count >= 1
    assert prrl_task.complete


# ------------------------------------------------------------------------------



from collections.abc import Callable
from crystal import resources
from crystal.util.test_mode import tests_are_running
from crystal.util.xos import is_kde_or_non_gnome, is_mac_os, is_windows
from typing import Protocol
import wx


def ShowModal(dialog: wx.Dialog) -> int:
    """
    Replacement for wx.Dialog.ShowModel() that works properly even when
    running automated tests.
    """
    is_running_tests = tests_are_running()  # cache
    
    if is_running_tests and isinstance(dialog, wx.MessageDialog):
        # A wx.MessageDialog opened with ShowModal cannot be interacted
        # with via wx.FindWindowByName() and similar functions on macOS.
        # So don't allow such a dialog to be shown while tests are running.
        raise AssertionError(
            f'Attempted to call ShowModal on wx.MessageDialog {dialog.Name!r} '
            f'while running an automated test, which would hang the test. '
            f'Please patch ShowModal to return an appropriate result.')
    
    # HACK: ShowModal sometimes hangs on macOS while running automated tests,
    #       as admitted by the wxPython test suite in test_dialog.py.
    #       So simulate its effect in a way that does NOT hang while running tests.
    if is_running_tests and is_mac_os():
        dialog.SetReturnCode(0)
        dialog.Show()
        
        loop = wx.GetApp().GetTraits().CreateEventLoop()
        with wx.EventLoopActivator(loop):
            while dialog.GetReturnCode() == 0:
                loop.Dispatch()
                if not dialog.IsShown() and dialog.GetReturnCode() == 0:
                    dialog.SetReturnCode(wx.ID_CANCEL)
        
        return dialog.GetReturnCode()
    else:
        return dialog.ShowModal()


class ShowModalFunc(Protocol):
    call_count: int
    def __call__(self, dialog: wx.Dialog) -> int: ...


def mocked_show_modal(
        dialog_name: str,
        return_code: int | Callable[[wx.Dialog], int],
        ) -> ShowModalFunc:
    """
    Creates a mocked version of ShowModal which verifies that the expected
    dialog was opened and returns the provided return code.
    """
    ShowModal: ShowModalFunc
    def ShowModal(dialog: wx.Dialog) -> int:  # type: ignore[no-redef]
        assert dialog_name == dialog.Name
        ShowModal.call_count += 1
        if callable(return_code):
            return return_code(dialog)
        else:
            return return_code
    ShowModal.call_count = 0
    return ShowModal


def position_dialog_initially(dialog: wx.Dialog) -> None:
    """
    Reposition the specified dialog by a default offset relative to
    its parent window.
    
    Different operating systems open dialogs at various positions
    relative to their parent window. This function standardizes
    the behavior across different operating systems. Native behavior:
    - macOS 10.14: (0, 0) offset
    - Windows 7: (200, 200) offset
    - Windows 10: (150, 200) offset
    - Ubuntu 22: (100, 100) offset
    """
    if dialog.Parent is None:
        return
    new_position = dialog.Parent.Position
    new_position.x += (150 // 2)
    new_position.y += (200 // 2)
    # NOTE: Linux sometimes ignores repositioning requests,
    #       especially if Wayland is being used rather than X11
    dialog.Position = new_position


def set_dialog_or_frame_icon_if_appropriate(tlw: wx.TopLevelWindow) -> None:
    # 1. Windows: Define app icon in the top-left corner
    # 2. KDE: Define app icon in the top-left corner and in the dock
    if is_windows() or is_kde_or_non_gnome():
        tlw.SetIcons(wx.IconBundle(resources.open_binary('appicon.ico')))


def CreateButtonSizer(
        parent: wx.Dialog,
        affirmative_id,
        cancel_id=wx.ID_CANCEL,
        ) -> wx.Sizer:
    """
    Has a similar effect as wx.Dialog.CreateButtonSizer() but supports any
    value for the `affirmative_id`.
    """
    sizer = wx.StdDialogButtonSizer()
    
    affirmative_button = wx.Button(parent, affirmative_id)
    affirmative_button.SetDefault()
    sizer.SetAffirmativeButton(affirmative_button)
    
    sizer.SetCancelButton(wx.Button(parent, cancel_id))
    
    sizer.Realize()
    return sizer



from __future__ import annotations

from collections.abc import Callable, Iterator
from contextlib import contextmanager
from crystal.browser.tasktree import TaskTreeNode
from crystal.model import Project
import crystal.task
from crystal.task import _is_scheduler_thread, scheduler_affinity, Task
from crystal.tests.util.controls import TreeItem
from crystal.tests.util.runner import bg_sleep
from crystal.tests.util.wait import (
    DEFAULT_WAIT_PERIOD, tree_has_children_condition,
    tree_has_no_children_condition, wait_for, wait_while, WaitTimedOut,
)
from crystal.tests.util.xthreading import bg_call_and_wait
from crystal.util.xthreading import fg_affinity
import re
import threading
from typing import List
from unittest.mock import patch
import wx

# ------------------------------------------------------------------------------
# Utility: Wait for Download

_MAX_DOWNLOAD_DURATION_PER_STANDARD_ITEM = (
    4 +  # fetch + parse time
    crystal.task.DELAY_BETWEEN_DOWNLOADS
) * 2.5  # fudge factor
_MAX_DOWNLOAD_DURATION_PER_LARGE_ITEM = (
    _MAX_DOWNLOAD_DURATION_PER_STANDARD_ITEM * 
    4
)
MAX_DOWNLOAD_DURATION_PER_ITEM = _MAX_DOWNLOAD_DURATION_PER_LARGE_ITEM

async def wait_for_download_to_start_and_finish(
        task_tree: wx.TreeCtrl,
        *, immediate_finish_ok: bool=False,
        stacklevel_extra: int=0
        ) -> None:
    period = DEFAULT_WAIT_PERIOD
    
    # Wait for start of download
    try:
        await wait_for(
            tree_has_children_condition(task_tree),
            timeout=4.0,  # 2.0s isn't long enough for Windows test runners on GitHub Actions
            message=lambda: f'Timed out waiting for download task to appear',
            stacklevel_extra=(1 + stacklevel_extra),
            screenshot_on_error=not immediate_finish_ok)
    except WaitTimedOut:
        if immediate_finish_ok:
            return
        else:
            raise
    
    # TODO: Eliminate fancy logic to determine `item_count` because it's no longer being used
    # 
    # Wait until download task is observed that says how many items are being downloaded
    item_count: int | None
    first_task_title_func = first_task_title_progression(task_tree)
    observed_titles = []  # type: List[str]
    did_start_download = False
    while True:
        download_task_title = first_task_title_func()
        if download_task_title is None:
            if did_start_download:
                # Didn't observe what the item count was
                # but we DID see evidence that a download actually started
                break
            if immediate_finish_ok:
                return
            raise AssertionError(
                'Download finished early without finding sub-resources. '
                'Did the download fail? '
                f'Task titles observed were: {observed_titles}')
        if download_task_title not in observed_titles:
            observed_titles.append(download_task_title)
        
        m = re.fullmatch(
            r'^(?:Downloading(?: group)?|Finding members of group): (.*?) -- (?:(\d+) of (?:at least )?(\d+) item\(s\)(?: -- .+)?|(.*))$',
            download_task_title)
        if m is None:
            raise AssertionError(
                f'Expected first task to be a download task but found task with title: '
                f'{download_task_title}')
        if m.group(4) is not None:
            if m.group(4) in [
                    'Waiting for response...',
                    'Parsing links...',
                    'Recording links...',
                    'Waiting before performing next request...']:
                did_start_download = True
            pass  # keep waiting
        else:
            did_start_download = True
            # NOTE: Currently unused. Just proving that we can calculate it.
            int(m.group(3))
            break
        
        await bg_sleep(period)
        continue
    assert did_start_download
    
    # Wait while downloading
    await wait_while(
        first_task_title_func,
        progress_timeout=MAX_DOWNLOAD_DURATION_PER_ITEM,
        progress_timeout_message=lambda: (
            f'Subresource download timed out after {MAX_DOWNLOAD_DURATION_PER_ITEM:.1f}s: '
            f'Stuck at status: {first_task_title_func()!r}'
        ),
        period=period,
    )
    
    # Ensure did finish downloading
    assert tree_has_no_children_condition(task_tree)()


def first_task_title_progression(task_tree: wx.TreeCtrl) -> Callable[[], str | None]:
    def first_task_title():
        root_ti = TreeItem.GetRootItem(task_tree)
        first_task_ti = root_ti.GetFirstChild()
        if first_task_ti is None:
            return None  # done
        return first_task_ti.Text
    return first_task_title


# ------------------------------------------------------------------------------
# Utility: Task -> TaskTreeNode

def ttn_for_task(task: Task) -> TaskTreeNode:
    for lis in task.listeners:
        if isinstance(lis, TaskTreeNode):
            return lis
    raise AssertionError(f'Unable to locate TaskTreeNode for {task!r}')

# ------------------------------------------------------------------------------
# Utility: Append Deferred Top-Level Tasks

@fg_affinity
@scheduler_affinity  # manual control of scheduler thread is assumed
def append_deferred_top_level_tasks(project: Project) -> None:
    """
    For any children whose appending to a project's RootTask was deferred by
    RootTask.append_child(), apply those appends now.
    
    Listeners that respond to the append of a child may take other actions
    directly after the append, such as completing the just-appended child.
    """
    project.root_task.append_deferred_top_level_tasks()
    
    # Postcondition
    assert len(project.root_task._children_to_add_soon) == 0


# ------------------------------------------------------------------------------
# Utility: Scheduler Manual Control

@contextmanager
def scheduler_disabled() -> Iterator[None]:
    """
    Context where the scheduler thread is disabled for any Projects that are opened.
    
    Projects opened before entering this context will continue to have active
    scheduler threads.
    """
    with patch('crystal.task.start_schedule_forever', lambda task: None), \
            scheduler_thread_context():
        yield


@contextmanager
def scheduler_thread_context(enabled: bool=True) -> Iterator[None]:
    """
    Context which executes its contents as if it was on a scheduler thread.
    
    For testing use only.
    """
    old_is_scheduler_thread = _is_scheduler_thread()  # capture
    setattr(threading.current_thread(), '_cr_is_scheduler_thread', enabled)
    try:
        assert enabled == _is_scheduler_thread()
        yield
    finally:
        setattr(threading.current_thread(), '_cr_is_scheduler_thread', old_is_scheduler_thread)


@scheduler_affinity  # manual control of scheduler thread is assumed
async def step_scheduler(
        project: Project,
        *, expect_done: bool=False,
        after_get: Callable[[], None] | None=None,
        ) -> bool:
    """
    Performs one unit of work from the scheduler.
    
    Returns whether the scheduler is done and no work is left.
    
    Arguments:
    * expect_done -- Whether it's expected that no work is left.
    """
    unit = project.root_task.try_get_next_task_unit()
    if after_get is not None:
        after_get()
    if unit is None:
        if not expect_done:
            raise AssertionError('Expected there to be no more tasks')
        return True
    else:
        if expect_done:
            raise AssertionError('Expected there to be at least one task remaining')
        await bg_call_and_wait(scheduler_thread_context()(unit))
        return False


@scheduler_affinity  # manual control of scheduler thread is assumed
def step_scheduler_now(
        project: Project,
        *, expect_done: bool=False,
        ) -> None:
    """
    Performs one unit of work from the scheduler.
    
    Arguments:
    * expect_done -- Whether it's expected that no work is left. Must be True.
    """
    if expect_done != True:
        raise ValueError('step_scheduler_now() only supports expect_done=True')
    unit = project.root_task.try_get_next_task_unit()
    assert expect_done
    assert unit is None


@scheduler_affinity  # manual control of scheduler thread is assumed
async def step_scheduler_until_done(project: Project) -> None:
    while True:
        unit = project.root_task.try_get_next_task_unit()  # step scheduler
        if unit is None:
            break
        await bg_call_and_wait(scheduler_thread_context()(unit))


def mark_as_complete(task: Task) -> None:
    assert not task.complete
    task.finish()
    assert task.complete


@contextmanager
@scheduler_affinity  # manual control of scheduler thread is assumed
def clear_top_level_tasks_on_exit(project: Project) -> Iterator[None]:
    try:
        yield
    finally:
        # Uncrash root task, if it was crashed
        project.root_task.crash_reason = None
        
        # Force root task children to complete
        project.root_task.append_deferred_top_level_tasks()
        for c in project.root_task.children:
            if not c.complete:
                mark_as_complete(c)
        
        # Clear root task children
        assert None == project.root_task.try_get_next_task_unit()
        assert len(project.root_task.children) == 0


# ------------------------------------------------------------------------------
